# -*- coding: utf-8 -*-

import select
from ..loop import IOLoop


class EpollLoop(IOLoop):
    def __init__(self):
        super(EpollLoop, self).__init__()
        self._epoll = select.epoll()

        self._poll = self._epoll.poll
        self._add_fd = self._epoll.register
        self._remove_fd = self._epoll.unregister
        self._modify_fd = self._epoll.modify

    def _poll(self, timeout):
        return self._epoll.poll(timeout)

    def _add_fd(self, fd, mode):
        self._epoll.register(fd, mode)

    def _remove_fd(self, fd):
        self._epoll.unregister(fd)

    def _modify_fd(self, fd, mode):
        self._epoll.modify(fd, mode)



# -*- coding: utf-8 -*-
# 2022/1/22
# create by: snower

import os
import string
import struct
import signal
import socket
import sevent

ascii_digits_letters = string.ascii_letters + string.digits
__SSL_CONTEXT_CACHE__ = {}

def config_signal():
    signal.signal(signal.SIGINT, lambda signum, frame: sevent.current().stop())
    signal.signal(signal.SIGTERM, lambda signum, frame: sevent.current().stop())

def get_address_environ(address, key):
    if address and isinstance(address, tuple):
        if isinstance(address[0], str):
            host_key = "".join([c if c in ascii_digits_letters else "_" for c in address[0]]).upper()
            if len(address) >= 2:
                value = os.environ.get("%s_%s_%s" % (key, host_key, address[1]))
                if value is not None:
                    return value
        if len(address) >= 2:
            value = os.environ.get("%s_%s" % (key, address[1]))
            if value is not None:
                return value
    return os.environ.get(key)

def config_ssl_context(address, context):
    import ssl
    versions = get_address_environ(address, "SEVENT_HELPERS_SSL_VERSIONS")
    if versions:
        versions = set(versions.replace(".", "_").split(","))
        for version in ("SSLv2", "SSLv3", "TLSv1", "TLSv1_1", "TLSv1_2", "TLSv1_3"):
            if hasattr(ssl, "OP_NO_%s" % version) and version not in versions:
                context.options |= getattr(ssl, "OP_NO_%s" % version)
    maximum_version = get_address_environ(address, "SEVENT_HELPERS_SSL_MAXIMUM_VERSION")
    if maximum_version and hasattr(ssl.TLSVersion, maximum_version.replace(".", "_")):
        context.maximum_version = getattr(ssl.TLSVersion, maximum_version.replace(".", "_"))
    minimum_version = get_address_environ(address, "SEVENT_HELPERS_SSL_MINIMUM_VERSION")
    if minimum_version and hasattr(ssl.TLSVersion, maximum_version.replace(".", "_")):
        context.minimum_version = getattr(ssl.TLSVersion, maximum_version.replace(".", "_"))
    alpn_protocols = get_address_environ(address, "SEVENT_HELPERS_SSL_ALPN_PROTOCOLS")
    if alpn_protocols:
        context.set_alpn_protocols([alpn_protocol for alpn_protocol in alpn_protocols.split(",")])
    npn_protocols = get_address_environ(address, "SEVENT_HELPERS_SSL_NPN_PROTOCOLS")
    if npn_protocols:
        context.set_npn_protocols([npn_protocol for npn_protocol in npn_protocols.split(",")])
    ciphers = get_address_environ(address, "SEVENT_HELPERS_SSL_CIPHERS")
    if ciphers:
        context.set_ciphers(ciphers)
    ecdh_curve = get_address_environ(address, "SEVENT_HELPERS_SSL_ECDH_CURVE")
    if ecdh_curve:
        context.set_ecdh_curve(ecdh_curve)
    disable_ticket = get_address_environ(address, "SEVENT_HELPERS_SSL_DISABLE_TICKET")
    if disable_ticket:
        context.options |= ssl.OP_NO_TICKET
    else:
        context.options &= ~ssl.OP_NO_TICKET

def create_server(address, *args, **kwargs):
    if not isinstance(address, (tuple, str)):
        address = tuple(address)
    if "pipe" in address:
        server = sevent.pipe.PipeServer()
    else:
        if address in __SSL_CONTEXT_CACHE__:
            context = __SSL_CONTEXT_CACHE__[address]
            server = sevent.ssl.SSLServer(context)
        else:
            ssl_certificate_file = get_address_environ(address, "SEVENT_HELPERS_SSL_CERTIFICATE_FILE")
            ssl_certificate_key_file = get_address_environ(address, "SEVENT_HELPERS_SSL_CERTIFICATE_KEY_FILE")
            if ssl_certificate_file and ssl_certificate_key_file:
                import ssl
                context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
                context.load_cert_chain(certfile=ssl_certificate_file, keyfile=ssl_certificate_key_file)
                ssl_ca_file = get_address_environ(address, "SEVENT_HELPERS_SSL_CA_FILE")
                if ssl_ca_file:
                    if ssl_ca_file == "-":
                        context.load_default_certs(ssl.Purpose.CLIENT_AUTH)
                    else:
                        context.load_verify_locations(cafile=ssl_ca_file)
                if get_address_environ(address, "SEVENT_HELPERS_SSL_VERIFY_OPTIONAL"):
                    context.verify_mode = ssl.CERT_OPTIONAL
                elif get_address_environ(address, "SEVENT_HELPERS_SSL_VERIFY_REQUIRED"):
                    context.verify_mode = ssl.CERT_REQUIRED
                config_ssl_context(address, context)
                __SSL_CONTEXT_CACHE__[address] = context
                server = sevent.ssl.SSLServer(context)
            else:
                server = sevent.tcp.Server()
    server.enable_reuseaddr()
    server.listen(address, *args, **kwargs)
    setattr(server, "address", address)
    return server

def create_socket_ssl_context(address, ssl_ca_file):
    import ssl
    context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH, cafile=ssl_ca_file if ssl_ca_file != "-" else None)
    ssl_certificate_file = get_address_environ(address, "SEVENT_HELPERS_SSL_CERTIFICATE_FILE")
    ssl_certificate_key_file = get_address_environ(address, "SEVENT_HELPERS_SSL_CERTIFICATE_KEY_FILE")
    if ssl_certificate_file and ssl_certificate_key_file:
        context.load_cert_chain(certfile=ssl_certificate_file, keyfile=ssl_certificate_key_file)
    if get_address_environ(address, "SEVENT_HELPERS_SSL_INSECURE"):
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
    else:
        if get_address_environ(address, "SEVENT_HELPERS_SSL_VERIFY_NONE"):
            context.verify_mode = ssl.CERT_NONE
        elif get_address_environ(address, "SEVENT_HELPERS_SSL_VERIFY_OPTIONAL"):
            context.verify_mode = ssl.CERT_OPTIONAL
        if get_address_environ(address, "SEVENT_HELPERS_SSL_VERIFY_HOSTNAME_NONE"):
            context.check_hostname = False
    config_ssl_context(address, context)
    return context

def create_socket(address):
    if not isinstance(address, (tuple, str)):
        address = tuple(address)
    if "pipe" in address:
        if isinstance(address, (tuple, list)):
            pipe_address = "pipe#%s" % (address[1] if len(address) >= 2 else address[-1])
        elif not isinstance(address, str):
            pipe_address = "pipe#%s" % address
        else:
            pipe_address = address
        if pipe_address in sevent.pipe.PipeServer._bind_servers:
            conn = sevent.pipe.PipeSocket()
        else:
            if address in __SSL_CONTEXT_CACHE__:
                context = __SSL_CONTEXT_CACHE__[address]
                conn = sevent.ssl.SSLSocket(context=context, server_hostname=address[0] if address and isinstance(address, tuple) else None)
            else:
                use_ssl_socket = get_address_environ(address, "SEVENT_HELPERS_SSL_SOCKET")
                ssl_ca_file = get_address_environ(address, "SEVENT_HELPERS_SSL_CA_FILE")
                if use_ssl_socket or ssl_ca_file:
                    context = create_socket_ssl_context(address, ssl_ca_file)
                    __SSL_CONTEXT_CACHE__[address] = context
                    conn = sevent.ssl.SSLSocket(context=context, server_hostname=address[0] if address and isinstance(address, tuple) else None)
                else:
                    conn = sevent.tcp.Socket()
    else:
        if address in __SSL_CONTEXT_CACHE__:
            context = __SSL_CONTEXT_CACHE__[address]
            conn = sevent.ssl.SSLSocket(context=context, server_hostname=address[0] if address and isinstance(address, tuple) else None)
        else:
            use_ssl_socket = get_address_environ(address, "SEVENT_HELPERS_SSL_SOCKET")
            ssl_ca_file = get_address_environ(address, "SEVENT_HELPERS_SSL_CA_FILE")
            if use_ssl_socket or ssl_ca_file:
                context = create_socket_ssl_context(address, ssl_ca_file)
                __SSL_CONTEXT_CACHE__[address] = context
                conn = sevent.ssl.SSLSocket(context=context, server_hostname=address[0] if address and isinstance(address, tuple) else None)
            else:
                conn = sevent.tcp.Socket()
    conn.enable_nodelay()
    return conn

def format_data_len(date_len):
    if date_len < 1024:
        return "%dB" % date_len
    elif date_len < 1024*1024:
        return "%.3fK" % (date_len/1024.0)
    elif date_len < 1024*1024*1024:
        return "%.3fM" % (date_len/(1024.0*1024.0))
    elif date_len < 1024*1024*1024*1024:
        return "%.3fG" % (date_len/(1024.0*1024.0*1024.0))
    return "%.3fT" % (date_len/(1024.0*1024.0*1024.0*1024.0))

def is_subnet(ip, subnet):
    try:
        ip = struct.unpack("!I", socket.inet_pton(socket.AF_INET, ip))[0]
        if isinstance(subnet[0], tuple) or isinstance(subnet[1], tuple):
            return False
        return (ip & subnet[1]) == (subnet[0] & subnet[1])
    except:
        ip = (struct.unpack("!QQ", socket.inet_pton(socket.AF_INET6, ip)))
        if not isinstance(subnet[0], tuple) or len(subnet[0]) != 2 or not isinstance(subnet[1], tuple) or len(subnet[1]) != 2:
            return False
        return ((ip[0] & subnet[1][0]) == (subnet[0][0] & subnet[1][0])) and ((ip[1] & subnet[1][1]) == (subnet[0][1] & subnet[1][1]))


# -*- coding: utf-8 -*-
# 2020/5/8
# create by: snower

import greenlet
from ..utils import get_logger


def warp_coroutine(BaseEventEmitter):
    class EventEmitter(BaseEventEmitter):
        def on(self, event_name, callback):
            if callback.__code__.co_flags & 0x80 == 0:
                return BaseEventEmitter.on(self, event_name, callback)

            def run_async_fuc(*args, **kwargs):
                def run():
                    try:
                        g = callback(*args, **kwargs)
                        g.send(None)
                        while True:
                            g.send(None)
                    except StopIteration:
                        pass
                    except Exception as e:
                        if isinstance(e, (KeyboardInterrupt, SystemError)):
                            raise e
                        get_logger().exception("error when calling callback:%s", e)
                child_gr = greenlet.greenlet(run)
                return child_gr.switch()
            BaseEventEmitter.on(self, event_name, run_async_fuc)

        def once(self, event_name, callback):
            if callback.__code__.co_flags & 0x80 == 0:
                return BaseEventEmitter.once(self, event_name, callback)

            def run_async_fuc(*args, **kwargs):
                def run():
                    try:
                        g = callback(*args, **kwargs)
                        g.send(None)
                        while True:
                            g.send(None)
                    except StopIteration:
                        pass
                    except Exception as e:
                        if isinstance(e, (KeyboardInterrupt, SystemError)):
                            raise e
                        get_logger().exception("error when calling callback:%s", e)
                child_gr = greenlet.greenlet(run)
                return child_gr.switch()
            BaseEventEmitter.once(self, event_name, run_async_fuc)

    return EventEmitter



# -*- coding: utf-8 -*-

import select
import time
import heapq
import threading
from collections import defaultdict
from .waker import Waker
from .utils import is_py3, get_logger

''' You can only use instance(). Don't create a Loop() '''

_thread_local = threading.local()
_thread_local._sevent_ioloop = None
_ioloop_lock = threading.RLock()
_ioloop_cls = None
_ioloop = None
_mul_ioloop = False


def instance():
    global _ioloop_cls, _ioloop, _mul_ioloop
    try:
        if _thread_local._sevent_ioloop is not None:
            return _thread_local._sevent_ioloop
    except AttributeError:
        pass

    with _ioloop_lock:
        try:
            if _thread_local._sevent_ioloop is not None:
                return _thread_local._sevent_ioloop
        except AttributeError:
            pass

        if 'epoll' in select.__dict__:
            from .impl import epoll_loop
            get_logger().debug('using epoll')
            _ioloop_cls = epoll_loop.EpollLoop
        elif 'kqueue' in select.__dict__:
            from .impl import kqueue_loop
            get_logger().debug('using kqueue')
            _ioloop_cls = kqueue_loop.KqueueLoop
        else:
            from .impl import select_loop
            get_logger().debug('using select')
            _ioloop_cls = select_loop.SelectLoop

        _thread_local._sevent_ioloop = _ioloop_cls()
        if _ioloop is None:
            _ioloop = _thread_local._sevent_ioloop
        else:
            _mul_ioloop = True
        return _thread_local._sevent_ioloop


def current():
    if not _mul_ioloop:
        return _ioloop
    try:
        return _thread_local._sevent_ioloop
    except AttributeError:
        return _ioloop


# these values are defined as the same as poll
MODE_NULL = 0x00
MODE_IN = 0x01
MODE_OUT = 0x04
MODE_ERR = 0x08
MODE_HUP = 0x10
MODE_NVAL = 0x20


class TimeoutHandler(object):
    def __init__(self, callback, deadline, args, kwargs):
        '''deadline here is absolute timestamp'''
        self.callback = callback
        self.deadline = deadline
        self.args = args
        self.kwargs = kwargs
        self.canceled = False

    def __cmp__(self, other):
        return (self.deadline > other.deadline) - (self.deadline < other.deadline)

    def __eq__(self, other):
        return self.deadline == other.deadline

    def __gt__(self, other):
        return self.deadline > other.deadline

    def __lt__(self, other):
        return self.deadline < other.deadline

    def __ge__(self, other):
        return self.deadline >= other.deadline

    def __le__(self, other):
        return self.deadline <= other.deadline

    def __ne__(self, other):
        return self.deadline != other.deadline

    def __call__(self):
        self.callback(*self.args, **self.kwargs)


class IOLoop(object):
    def __init__(self):
        self._handlers = []
        self._run_handlers = []
        self._timeout_handlers = []
        self._fd_handlers = defaultdict(list)
        self._stopped = False
        self._waker = Waker()

    def _poll(self, timeout):
        raise NotImplementedError()

    def _add_fd(self, fd, mode):
        raise NotImplementedError()

    def _remove_fd(self, fd):
        raise NotImplementedError()

    def _modify_fd(self, fd, mode):
        raise NotImplementedError()

    def add_fd(self, fd, mode, callback):
        handlers = self._fd_handlers[fd]
        new_handlers = []
        if not handlers:
            new_handlers.append((callback, fd, mode))
            self._add_fd(fd, mode)
        else:
            new_mode = MODE_NULL
            for hcallback, hfd, hmode in handlers:
                if hcallback != callback:
                    new_mode |= hmode
                    new_handlers.append((hcallback, hfd, hmode))
            new_handlers.append((callback, fd, mode))
            new_mode |= mode
            self._modify_fd(fd, new_mode)
        self._fd_handlers[fd] = new_handlers
        return True

    def update_fd(self, fd, mode, callback):
        handlers = self._fd_handlers[fd]
        if not handlers:
            return False
        new_handlers = []
        new_mode = MODE_NULL
        for hcallback, hfd, hmode in handlers:
            if hcallback == callback:
                new_mode |= mode
                new_handlers.append((hcallback, hfd, mode))
            else:
                new_mode |= hmode
                new_handlers.append((hcallback, hfd, hmode))
        self._modify_fd(fd, new_mode)
        self._fd_handlers[fd] = new_handlers
        return True

    def remove_fd(self, fd, callback):
        handlers = self._fd_handlers[fd]
        if not handlers:
            return False
        if len(handlers) == 1:
            if handlers[0][0] == callback:
                self._remove_fd(fd)
                del self._fd_handlers[fd]
        else:
            new_handlers = []
            new_mode = MODE_NULL
            for hcallback, hfd, hmode in handlers:
                if hcallback != callback:
                    new_mode |= hmode
                    new_handlers.append((hcallback, hfd, hmode))
            self._modify_fd(fd, new_mode)
            self._fd_handlers[fd] = new_handlers
        return True

    def clear_fd(self, fd):
        if fd in self._fd_handlers:
            if self._fd_handlers[fd]:
                del self._fd_handlers[fd]
                self._remove_fd(fd)
                return True
            del self._fd_handlers[fd]
        return False

    def start(self):
        self.add_fd(self._waker.fileno(), MODE_IN, self._waker.consume)

        while not self._stopped:
            timeout = 3600

            if self._timeout_handlers:
                cur_time = time.time()
                if self._timeout_handlers[0].deadline <= cur_time:
                    while self._timeout_handlers:
                        handler = self._timeout_handlers[0]
                        if handler.canceled:
                            heapq.heappop(self._timeout_handlers)
                        elif handler.deadline <= cur_time:
                            heapq.heappop(self._timeout_handlers)
                            try:
                                handler.callback(*handler.args, **handler.kwargs)
                            except Exception as e:
                                if isinstance(e, (KeyboardInterrupt, SystemError)):
                                    raise e
                                get_logger().exception("loop callback timeout error:%s", e)
                        elif self._handlers:
                            timeout = 0
                            break
                        else:
                            timeout = self._timeout_handlers[0].deadline - cur_time
                            break
                elif self._handlers:
                    timeout = 0
                else:
                    timeout = self._timeout_handlers[0].deadline - cur_time
            elif self._handlers:
                timeout = 0

            fds_ready = self._poll(timeout)
            for fd, mode in fds_ready:
                for hcallback, hfd, hmode in self._fd_handlers[fd]:
                    if hmode & mode != 0:
                        try:
                            hcallback()
                        except Exception as e:
                            if isinstance(e, (KeyboardInterrupt, SystemError)):
                                raise e
                            get_logger().exception("loop callback error:%s", e)

            # call handlers without fd
            self._handlers, self._run_handlers = self._run_handlers, self._handlers
            for callback, args, kwargs in self._run_handlers:
                try:
                    callback(*args, **kwargs)
                except Exception as e:
                    if isinstance(e, (KeyboardInterrupt, SystemError)):
                        raise e
                    get_logger().exception("loop callback error:%s", e)
            self._run_handlers = []

    def stop(self):
        self._stopped = True
        self._waker.wake()

    def add_async(self, callback, *args, **kwargs):
        self._handlers.append((callback, args, kwargs))

    def add_async_safe(self, callback, *args, **kwargs):
        self._handlers.append((callback, args, kwargs))
        self._waker.wake()

    def add_timeout(self, timeout, callback, *args, **kwargs):
        handler = TimeoutHandler(callback, time.time() + timeout, args, kwargs)
        heapq.heappush(self._timeout_handlers, handler)
        return handler

    def cancel_timeout(self, handler):
        if handler.__class__ is TimeoutHandler:
            handler.callback = None
            handler.args = None
            handler.kwargs = None
            handler.canceled = True
        else:
            try:
                self._timeout_handlers.remove(handler)
                heapq.heapify(self._timeout_handlers)
            except ValueError:
                pass

        while self._timeout_handlers:
            if not self._timeout_handlers[0].canceled:
                break
            heapq.heappop(self._timeout_handlers)

    def wakeup(self, *args, **kwargs):
        if args and callable(args[0]):
            self.add_async(args[0], *args[1:], **kwargs)
        self._waker.wake()


if is_py3:
    from .coroutines.loop import warp_coroutine
    IOLoop = warp_coroutine(IOLoop)



# -*- coding: utf-8 -*-
# 2020/7/10
# create by: snower

import sys
import time
import struct
import random
import traceback
import logging
import argparse
from collections import deque
import threading
import socket
import sevent
from .utils import create_server, create_socket, config_signal, format_data_len, is_subnet

BYTES_MAP = {"B": 1, "K": 1024, "M": 1024*1024, "G": 1024*1024*1024, "T": 1024*1024*1024*1024}

def host_parse(host):
    hosts, subnet, cs = [], ["", 0], []
    is_brackets, is_subnet = False, False
    for c in host:
        if c == ":":
            if not is_brackets and not is_subnet:
                hosts.append("".join(cs))
                cs = []
            else:
                cs.append(c)
        elif c == "[":
            is_brackets = True
        elif c == "]":
            is_brackets = False
        elif c == "|":
            hosts.append("".join(cs))
            cs = []
            is_subnet = True
        elif c == "/":
            subnet[0] = "".join(cs)
            cs = []
        else:
            cs.append(c)

    if is_subnet:
        if subnet[0]:
            subnet[1] = int("".join(cs) if cs else "32")
        else:
            subnet = ["".join(cs), 32]
    else:
        hosts.append("".join(cs))
        subnet = ["0.0.0.0", 0]

    if len(hosts) == 2:
        if hosts[0].isdigit() and hosts[1].isdigit():
            hosts = [("0.0.0.0", int(hosts[0])), ("127.0.0.1", int(hosts[1]))]
        elif hosts[0] == "":
            hosts = [("0.0.0.0", 0), ("127.0.0.1", int(hosts[1]))]
        else:
            hosts = [("0.0.0.0", 0), (hosts[0]), int(hosts[1])]
    elif len(hosts) == 3:
        if hosts[0].isdigit():
            hosts = [("0.0.0.0", int(hosts[0])), (hosts[1], int(hosts[2]))]
        elif hosts[0] == "":
            hosts = [("0.0.0.0", 0), (hosts[1], int(hosts[2]))]
        elif hosts[2].isdigit():
            hosts = [(hosts[0], int(hosts[1])), ("127.0.0.1", int(hosts[2]))]
        else:
            hosts = [(hosts[0], int(hosts[1])), ("127.0.0.1", 0)]
    elif len(hosts) == 4:
        hosts = [(hosts[0], int(hosts[1])), (hosts[2], int(hosts[3]))]
    else:
        raise ValueError(u"host error %s", host)

    try:
        subnet[0] = struct.unpack("!I", socket.inet_pton(socket.AF_INET, subnet[0]))[0]
        subnet[1] = ~ (0xffffffff >> subnet[1])
    except:
        subnet[0] = (struct.unpack("!QQ", socket.inet_pton(socket.AF_INET6, subnet[0])))
        subnet[1] = (~ (0xffffffffffffffff >> min(subnet[1], 64)), ~ (0xffffffffffffffff >> max(subnet[1] - 64, 0)))
    return hosts, subnet

def warp_write(conn, status, key):
    origin_write = conn.write
    def _(data):
        status[key] += len(data)
        return origin_write(data)
    return _

def warp_speed_limit_write(conn, status, key):
    conn_id = id(conn)
    origin_write = conn.write
    origin_end = conn.end
    speed_limiter = status["speed_limiter"]
    current_speed_key = "recv_current_speed" if key == "recv_len" else "send_current_speed"
    end_key = "recv_is_end" if key == "recv_len" else "send_is_end"
    status[end_key] = False
    buffer = sevent.buffer.Buffer()

    def warp_end():
        if conn_id not in speed_limiter.buffers:
            return origin_end()
        status[end_key] = True
    conn.end = warp_end

    def speed_write(data, speed):
        dl = len(data)
        if dl > speed:
            if speed <= 0:
                if speed_limiter.current_speed <= 0:
                    return
                speed = min(speed_limiter.current_speed, speed_limiter.speed)
                if dl > speed:
                    status[key] += buffer.fetch(data, speed)
                    speed_limiter.current_speed -= speed
                else:
                    status[key] += dl
                    speed_limiter.current_speed -= dl
            else:
                status[key] += buffer.fetch(data, speed)
            try:
                return origin_write(buffer)
            except sevent.tcp.SocketClosed:
                speed_limiter.buffers.pop(conn_id, None)
                if status[end_key]:
                    status[end_key] = False
                    sevent.current().add_async(origin_end)
                return False

        if not data:
            speed_limiter.buffers.pop(conn_id, None)
            if status[end_key]:
                status[end_key] = False
                sevent.current().add_async(origin_end)
            return True

        status[key] += dl
        try:
            return origin_write(data)
        except sevent.tcp.SocketClosed:
            return False

    def _(data):
        if conn_id in speed_limiter.buffers:
            current_speed = status[current_speed_key]
            if current_speed <= 0:
                return

            if speed_limiter.global_speed:
                if speed_limiter.current_speed <= 0:
                    return

                dl = len(data)
                speed = min(speed_limiter.current_speed, current_speed)
                if dl > speed:
                    status[key] += buffer.fetch(data, speed)
                    speed_limiter.current_speed -= speed
                    status[current_speed_key] -= speed
                    return origin_write(buffer)

                status[key] += dl
                speed_limiter.current_speed -= dl
                status[current_speed_key] -= dl
                return origin_write(data)

            dl = len(data)
            if dl > current_speed:
                status[key] += buffer.fetch(data, current_speed)
                status[current_speed_key] = 0
                return origin_write(buffer)

            status[key] += dl
            status[current_speed_key] -= dl
            return origin_write(data)

        speed_limiter.buffers[conn_id] = (data, speed_write)
        if not speed_limiter.is_running:
            speed_limiter.is_running = True
            sevent.current().call_async(speed_limiter.loop)

        dl = len(data)
        if speed_limiter.global_speed:
            if speed_limiter.current_speed <= 0:
                status[current_speed_key] = 0
                return
            speed = min(speed_limiter.current_speed, speed_limiter.speed)
            if dl > speed:
                status[key] += buffer.fetch(data, speed)
                speed_limiter.current_speed -= speed
                status[current_speed_key] = 0
                return origin_write(buffer)

            status[key] += dl
            speed_limiter.current_speed -= dl
            status[current_speed_key] = speed_limiter.speed - dl
            return origin_write(data)

        if dl > speed_limiter.speed:
            status[key] += buffer.fetch(data, speed_limiter.speed)
            status[current_speed_key] = 0
            return origin_write(buffer)

        status[key] += dl
        status[current_speed_key] = speed_limiter.speed - dl
        return origin_write(data)
    return _

def warp_delay_write(delayer, warp_write_func):
    def _(conn, status, key):
        origin_write = warp_write_func(conn, status, key)
        origin_end = conn.end
        buffer = sevent.buffer.Buffer()
        key = "delay_%s_buffer_len" % id(buffer)
        status[key] = 0
        end_key = "delay_recv_is_end" if key == "recv_len" else "delay_send_is_end"
        status[end_key] = False

        def warp_end():
            if status[key] <= 0:
                return origin_end()
            status[end_key] = True
        conn.end = warp_end

        def delay_write(data, data_len):
            status[key] -= buffer.fetch(data, data_len)
            try:
                return origin_write(buffer)
            except sevent.tcp.SocketClosed:
                return False
            finally:
                if not data and status[end_key]:
                    status[end_key] = False
                    sevent.current().add_async(origin_end)

        def __(data):
            data_len = max(len(data) - status[key], 0)
            if delayer.delay:
                delayer.queues.append((time.time() + delayer.delay, data, data_len, delay_write))
            else:
                delayer.queues.append((time.time() + random.randint(*delayer.rdelays) / 1000000.0,
                                       data, data_len, delay_write))
            status[key] += data_len
            if delayer.is_running:
                return
            delayer.is_running = True
            sevent.current().call_async(delayer.loop)
        return __
    return _

def warp_mirror_write(mirror_host, mirror_header, warp_write_func):
    try:
        mirror_host, subnet = host_parse(mirror_host)
        up_address, down_address = mirror_host[0], mirror_host[1]
        if up_address[0] == "0.0.0.0":
            up_address = ("127.0.0.1", up_address[1])
        if down_address[0] == "0.0.0.0":
            down_address = ("127.0.0.1", down_address[1])
    except ValueError:
        logging.error("mirror address error %s", mirror_host)
        return warp_write_func
    mirror_header = mirror_header.replace("{", "%(").replace("}", ")s")

    def warp_up_conn_write(conn, status, key):
        origin_write = conn.write
        up_buffer = sevent.buffer.Buffer()

        def up_conn_write(data):
            try:
                if "mirror_up_conn" not in status:
                    if up_address == down_address and status.get("mirror_down_conn"):
                        status['mirror_up_conn'] = status.get("mirror_down_conn")
                    else:
                        mirror_conn = create_socket(up_address)
                        mirror_conn.connect(up_address, 5)
                        mirror_conn.on_connect(lambda s: mirror_conn.end() if conn._state == sevent.tcp.STATE_CLOSED else None)
                        mirror_conn.on_data(lambda s, b: b.read())
                        conn.on_close(lambda s: mirror_conn.end() if mirror_conn._state != sevent.tcp.STATE_CONNECTING else None)
                        status['mirror_up_conn'] = mirror_conn
                        try:
                            if mirror_header:
                                up_buffer.write((mirror_header % status["mirror_variables"]).encode("utf-8"))
                        except:
                            pass
                    logging.info("mirror up copy to %s:%s", up_address[0], up_address[1])
                up_buffer.copyfrom(data)
                status['mirror_up_conn'].write(up_buffer)
            except:
                pass
            return origin_write(data)
        return up_conn_write

    def warp_down_conn_write(conn, status, key):
        origin_write = conn.write
        down_buffer = sevent.buffer.Buffer()

        def down_conn_write(data):
            try:
                if "mirror_down_conn" not in status:
                    if up_address == down_address and status.get("mirror_up_conn"):
                        status['mirror_down_conn'] = status.get("mirror_up_conn")
                    else:
                        mirror_conn = create_socket(down_address)
                        mirror_conn.connect(down_address, 5)
                        mirror_conn.on_connect(lambda s: mirror_conn.end() if conn._state == sevent.tcp.STATE_CLOSED else None)
                        mirror_conn.on_data(lambda s, b: b.read())
                        conn.on_close(lambda s: mirror_conn.end() if mirror_conn._state != sevent.tcp.STATE_CONNECTING else None)
                        status['mirror_down_conn'] = mirror_conn
                        try:
                            if mirror_header:
                                down_buffer.write((mirror_header % status["mirror_variables"]).encode("utf-8"))
                        except Exception as e:
                            pass
                    logging.info("mirror down copy to %s:%s", down_address[0], down_address[1])
                down_buffer.copyfrom(data)
                status['mirror_down_conn'].write(down_buffer)
            except:
                pass
            return origin_write(data)
        return down_conn_write

    def _(conn, status, key):
        if mirror_header and "mirror_variables" not in status:
            status["mirror_variables"] = {"conn_id": "", "from_host": "", "from_port": 0, "to_host": "", "to_port": 0}
        if key == "send_len":
            if "mirror_subnet" not in status or status["mirror_subnet"]:
                if up_address and len(up_address) >= 2 and up_address[1] > 0:
                    conn.write = warp_up_conn_write(conn, status, key)
                if mirror_header:
                    status["mirror_variables"]["to_host"] = conn.address[0]
                    status["mirror_variables"]["to_port"] = conn.address[1]
            else:
                status.pop("mirror_variables", None)
        if key == "recv_len":
            if subnet and not is_subnet(conn.address[0], subnet):
                status["mirror_subnet"] = False

            if "mirror_subnet" not in status or status["mirror_subnet"]:
                if down_address and len(down_address) >= 2 and down_address[1] > 0:
                    conn.write = warp_down_conn_write(conn, status, key)
                if mirror_header:
                    status["mirror_variables"]["conn_id"] = id(conn)
                    status["mirror_variables"]["from_host"] = conn.address[0]
                    status["mirror_variables"]["from_port"] = conn.address[1]
            else:
                status.pop("mirror_variables", None)
        return warp_write_func(conn, status, key)
    return _

async def tcp_forward(conns, conn, forward_address, subnet, status):
    start_time = time.time()
    conn.write, pconn = warp_write(conn, status, "recv_len"), None

    try:
        conn.enable_nodelay()
        pconn = create_socket(forward_address)
        await pconn.connectof(forward_address)
        pconn.write = warp_write(pconn, status, "send_len")
        logging.info("tcp forward connected %s:%d -> %s:%d", conn.address[0], conn.address[1], forward_address[0], forward_address[1])
        await pconn.linkof(conn)
    except sevent.errors.SocketClosed:
        pass
    except Exception as e:
        logging.info("tcp forward error %s:%d -> %s:%d %s %.2fms\r%s", conn.address[0], conn.address[1],
                     forward_address[0], forward_address[1], e, (time.time() - start_time) * 1000, traceback.format_exc())
        return
    finally:
        conn.close()
        if pconn: pconn.close()
        conns.pop(id(conn), None)

    logging.info("tcp forward closed %s:%d -> %s:%d %s %s %.2fms", conn.address[0], conn.address[1],
                 forward_address[0], forward_address[1], format_data_len(status["send_len"]), format_data_len(status["recv_len"]),
                 (time.time() - start_time) * 1000)

async def tcp_forward_server(conns, server, forward_hosts, speed_limiter):
    while True:
        conn = await server.accept()
        conn_ip, forward_address, subnet = conn.address[0], None, None
        for a, s in forward_hosts:
            if is_subnet(conn_ip, s):
                forward_address, subnet = a, s
                break

        if not forward_address:
            conn.close()
            logging.info("tcp forward subnet fail %s:%d", conn.address[0], conn.address[1])
            return

        status = {"recv_len": 0, "send_len": 0, "last_time": time.time(), "check_recv_len": 0, "check_send_len": 0, "speed_limiter": speed_limiter}
        sevent.current().call_async(tcp_forward, conns, conn, forward_address, subnet, status)
        conns[id(conn)] = (conn, status)

async def check_timeout(conns, timeout):
    def run_check():
        while True:
            try:
                now = time.time()
                for conn_id, (conn, status) in list(conns.items()):
                    if status['check_recv_len'] != status['recv_len'] or status['check_send_len'] != status['send_len']:
                        status["check_recv_len"] = status["recv_len"]
                        status["check_send_len"] = status["send_len"]
                        status['last_time'] = now
                        continue

                    if now - status['last_time'] >= timeout:
                        sevent.current().add_async_safe(conn.close)
                        conns.pop(conn_id, None)
            finally:
                time.sleep(min(float(timeout) / 2.0, 30))

    if timeout > 0:
        check_thread = threading.Thread(target=run_check)
        check_thread.daemon = True
        check_thread.start()
    await sevent.Future()

class Delayer(object):
    def __init__(self, delay, rdelays):
        self.delay = delay
        self.rdelays = rdelays
        self.is_running = False
        self.queues = deque()

    async def loop(self):
        try:
            while self.queues:
                now = time.time()
                timeout_time, data, data_len, callback = self.queues.popleft()
                if timeout_time > now:
                    await sevent.current().sleep(timeout_time - now)
                sevent.current().add_async(callback, data, data_len)
        finally:
            self.is_running = False

class SpeedLimiter(object):
    def __init__(self, speed, global_speed):
        self.speed = int(speed) or int(global_speed)
        self.global_speed = int(global_speed)
        self.current_speed = self.global_speed if self.global_speed else 0
        self.buffers = {}
        self.is_running = False
        self.loop = self.loop_global_speed if self.global_speed else self.loop_speed

    async def loop_speed(self):
        try:
            current_timestamp = time.time()
            await sevent.current().sleep(0.1)
            while self.buffers:
                try:
                    for _, (data, callback) in list(self.buffers.items()):
                        sevent.current().add_async(callback, data, self.speed)
                finally:
                    now = time.time()
                    sleep_time = 0.2 - (now - current_timestamp)
                    current_timestamp = now
                    await sevent.current().sleep(sleep_time)
        finally:
            self.is_running = False

    async def loop_global_speed(self):
        try:
            current_timestamp = time.time()
            await sevent.current().sleep(0.1)
            while self.buffers:
                try:
                    avg_speed = int(self.global_speed / len(self.buffers))
                    max_speed, over_speed = min(avg_speed, self.speed), 0
                    speed_buffers = []
                    for _, (data, callback) in self.buffers.items():
                        dl = len(data)
                        if max_speed > dl:
                            speed_buffers.append((data, dl, dl, callback))
                            over_speed += avg_speed - dl
                        else:
                            speed_buffers.append((data, dl, max_speed, callback))
                            over_speed += avg_speed - max_speed

                    for data, dl, speed, callback in speed_buffers:
                        if over_speed > 0 and dl > speed and speed < self.speed:
                            can_speed = min(min(dl, self.speed) - speed, over_speed)
                            speed += can_speed
                            over_speed -= can_speed
                        sevent.current().add_async(callback, data, speed)
                    self.current_speed = over_speed
                finally:
                    now = time.time()
                    sleep_time = 0.2 - (now - current_timestamp)
                    current_timestamp = now
                    await sevent.current().sleep(sleep_time)
        finally:
            self.current_speed = self.global_speed
            self.is_running = False

async def tcp_forward_servers(servers, timeout, speed, global_speed):
    conns, speed_limiter = {}, (SpeedLimiter(speed, global_speed) if speed or global_speed else None)
    for server, forward_hosts in servers:
        sevent.current().call_async(tcp_forward_server, conns, server, forward_hosts, speed_limiter)
    await check_timeout(conns, timeout)

def main(argv):
    global warp_write

    parser = argparse.ArgumentParser(description="tcp port forward")
    parser.add_argument('-L', dest='forwards', default=[], action="append", type=str,
                        help='forward host, accept format [[local_bind:]local_port:remote_host:remote_port]|[subnet], support muiti forward args (example: 0.0.0.0:80:127.0.0.1:8088 or 80:192.168.0.2:8088|192.168.0.0/24)')
    parser.add_argument('-t', dest='timeout', default=7200, type=int, help='no read/write timeout (default: 7200)')
    parser.add_argument('-s', dest='speed', default=0, type=lambda v: int(float(v[:-1]) * BYTES_MAP[v.upper()[-1]]) \
        if v and v.upper()[-1] in BYTES_MAP else int(float(v)), help='per connection speed limit byte, example: 1024 or 1M (default: 0 is unlimit), available units : B K M G T')
    parser.add_argument('-S', dest='global_speed', default=0, type=lambda v: int(float(v[:-1]) * BYTES_MAP[v.upper()[-1]]) \
        if v and v.upper()[-1] in BYTES_MAP else int(float(v)), help='global speed limit byte, example: 1024 or 1M (default: 0 is unlimit), available units : B K M G T')
    parser.add_argument('-d', dest='delay', default=0, type=lambda v: (float(v.split("-")[0]), float(v.split("-")[-1])) \
        if v and isinstance(v, str) and "-" in v else float(v), help='delay millisecond (default: 0 is not delay, example -d100 or -d100-200), the - between two numbers will be random delay')
    parser.add_argument('-M', dest='mirror_host', default="", type=str,
                        help='mirror host, accept format [[up_host:]up_port:[down_host]:down_port] (example: 0.0.0.0:80:127.0.0.1:8088 or :127.0.0.1:8088 or 127.0.0.1:8088: or 8088:8088)')
    parser.add_argument('-F', dest='mirror_header', default="", type=str,
                        help='mirror header, accept variables [from_host|from_port|to_host|to_port|conn_id] (example: "{conn_id}-{from_host}:{from_port}->{to_host}:{to_port}\\r\\n")')
    args = parser.parse_args(args=argv)
    config_signal()
    if not args.forwards:
        exit(0)

    forwards = {}
    for forward in args.forwards:
        hosts, subnet = host_parse(forward)
        if hosts[0] not in forwards:
            forwards[hosts[0]] = []
        forwards[hosts[0]].append((hosts[1], subnet))
    if not forwards:
        exit(0)

    if args.speed or args.global_speed:
        warp_write = warp_speed_limit_write

    if args.delay:
        warp_write = warp_delay_write(Delayer(0 if isinstance(args.delay, tuple) else float(args.delay) / 1000.0,
                                    (int(args.delay[0] * 1000), int(args.delay[1] * 1000)) if isinstance(args.delay, tuple) else None),
                                    warp_write)

    if args.mirror_host:
        warp_write = warp_mirror_write(args.mirror_host, args.mirror_header, warp_write)

    forward_servers = []
    for bind_address, forward_hosts in forwards.items():
        server = create_server(bind_address)
        forward_servers.append((server, sorted(forward_hosts,
                                               key=lambda x: x[1][1][0] * 0xffffffffffffffff + x[1][1][1] if isinstance(x[1][1], tuple) else x[1][1],
                                               reverse=True)))
        logging.info("port forward listen %s:%s", bind_address[0], bind_address[1])

    sevent.current().call_async(tcp_forward_servers, forward_servers, args.timeout,
                                int(args.speed / 10), int(args.global_speed / 10))

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)1.1s %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S', filemode='a+')
    try:
        main(sys.argv[1:])
        sevent.instance().start()
    except KeyboardInterrupt:
        exit(0)


# -*- coding: utf-8 -*-
# 2020/5/10
# create by: snower

import greenlet
from ..utils import get_logger

_PENDING = 'PENDING'
_CANCELLED = 'CANCELLED'
_FINISHED = 'FINISHED'


class CancelledError(Exception): pass


class InvalidStateError(Exception): pass


class Future(object):
    _state = _PENDING
    _result = None
    _exception = None
    _source_traceback = None
    _log_traceback = False

    def __init__(self):
        self._callbacks = []

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self._callbacks)

    def cancel(self):
        self._log_traceback = False
        if self._state != _PENDING:
            return False
        self._state = _CANCELLED
        self._schedule_callbacks()
        return True

    def _schedule_callbacks(self):
        callbacks = self._callbacks[:]
        if not callbacks:
            return

        self._callbacks[:] = []
        for callback in callbacks:
            try:
                callback(self)
            except Exception as e:
                get_logger().exception("future schedule callback error:%s", e)

    def cancelled(self):
        return self._state == _CANCELLED

    def done(self):
        return self._state != _PENDING

    def result(self):
        if self._state == _CANCELLED:
            raise CancelledError
        if self._state != _FINISHED:
            raise InvalidStateError('Result is not ready.')
        self._log_traceback = False
        if self._exception is not None:
            raise self._exception
        return self._result

    def exception(self):
        if self._state == _CANCELLED:
            raise CancelledError
        if self._state != _FINISHED:
            raise InvalidStateError('Exception is not set.')
        self._log_traceback = False
        return self._exception

    def add_done_callback(self, fn):
        if self._state != _PENDING:
            fn(self)
        else:
            self._callbacks.append(fn)

    def remove_done_callback(self, fn):
        filtered_callbacks = [f for f in self._callbacks if f != fn]
        removed_count = len(self._callbacks) - len(filtered_callbacks)
        if removed_count:
            self._callbacks[:] = filtered_callbacks
        return removed_count

    def set_result(self, result):
        if self._state != _PENDING:
            raise InvalidStateError('{}: {!r}'.format(self._state, self))
        self._result = result
        self._state = _FINISHED
        self._schedule_callbacks()

    def set_exception(self, exception):
        if self._state != _PENDING:
            raise InvalidStateError('{}: {!r}'.format(self._state, self))
        if isinstance(exception, type):
            exception = exception()
        if type(exception) is StopIteration:
            raise TypeError("StopIteration interacts badly with generators "
                            "and cannot be raised into a Future")
        self._exception = exception
        self._state = _FINISHED
        self._schedule_callbacks()

    def __next__(self):
        if not self.done():
            child_gr = greenlet.getcurrent()
            main = child_gr.parent
            assert main is not None, "must be running in async func"
            self._callbacks.append(lambda future: child_gr.switch())
            main.switch()

        result = self.result()
        e = StopIteration()
        e.value = result
        raise e

    def __iter__(self):
        return self

    __await__ = __iter__



# -*- coding: utf-8 -*-
# 2020/7/10
# create by: snower

import sys
import time
import re
import struct
import socket
import argparse
import logging
import traceback
import threading
import sevent
from .utils import create_server, create_socket, config_signal, format_data_len

def parse_allow_deny_host(host_name):
    try:
        if ":" in host_name[host_name.index("]") + 1]:
            host_name = host_name.replace("[", "\[").replace("]", "\]").replace(".", "\.").replace("*", ".+?")
        else:
            host_name = host_name.replace("[", "\[").replace("]", "\]").replace(".", "\.").replace("*", ".*?") + ":.*?"
    except:
        if ":" in host_name:
            host_name = host_name.replace("[", "\[").replace("]", "\]").replace(".", "\.").replace("*", ".+?")
        else:
            host_name = host_name.replace("[", "\[").replace("]", "\]").replace(".", "\.").replace("*", ".*?") + ":.*?"
    return re.compile("^" + host_name + "$")

def parse_allow_deny_host_filename(filename):
    host_names = set([])
    with open(filename, "r", encoding="utf-8") as fp:
        for line in fp:
            host_names.add(parse_allow_deny_host(line.strip()))
    return host_names

def check_allow_deny_domain(allow_host_name_res, deny_host_name_res, address):
    if not allow_host_name_res and not deny_host_name_res:
        return True
    check_host_name = "%s:%d" % (address[0], address[1])
    if allow_host_name_res:
        for allow_host_name_re in allow_host_name_res:
            if allow_host_name_re.match(check_host_name):
                return True
        return False
    if deny_host_name_res:
        for deny_host_name_re in deny_host_name_res:
            if deny_host_name_re.match(check_host_name):
                return False
        return True
    return True

def warp_write(conn, status, key):
    origin_write = conn.write
    def _(data):
        status[key] += len(data)
        return origin_write(data)
    return _

def http_protocol_parse_address(data, default_port=80):
    address = data.split(b":")
    if len(address) != 2:
        return address[0].decode("utf-8"), default_port
    return address[0].decode("utf-8"), int(address[1].decode("utf-8"))

async def http_protocol_parse(conn, buffer, allow_host_names=None, deny_host_names=None):
    data = buffer.read()
    index = data.find(b" ")
    method = data[:index].decode("utf-8")
    data = data[index + 1:]

    if method.lower() == "connect":
        index = data.find(b" ")
        host, port = http_protocol_parse_address(data[:index], 443)
        if not check_allow_deny_domain(allow_host_names, deny_host_names, (host, port)):
            await conn.send(b"HTTP/1.1 403 Forbidden\r\n\r\n")
            return False, host, port, b''
        await conn.send(b"HTTP/1.1 200 Connection Established\r\n\r\n")
        return True, host, port, b''

    if method.lower() in ("get", "post", "put", "options", "head", "delete", "patch"):
        data = data[7:]
        index = data.find(b"/")
        host, port = http_protocol_parse_address(data[:index], 80)
        if not check_allow_deny_domain(allow_host_names, deny_host_names, (host, port)):
            await conn.send(b"HTTP/1.1 403 Forbidden\r\n\r\n")
            return False, host, port, b''
        return True, host, port, method.encode("utf-8") + b' ' + data[index:]
    return False, '', 0, None

async def socks5_protocol_parse(conn, buffer, allow_host_names=None, deny_host_names=None):
    data = buffer.read()
    if len(data) < 8 or data[:3] != b"\x05\x01\x8e":
        is_socks5s, protocol = False, "socks5"
        if len(data) < 3 or data[0] != 5 or data[1] <= 0 or b'\x00' not in data[2:2 + data[1]]:
            return False, protocol, '', 0, None
        await conn.send(b'\x05\x00')
        buffer = await conn.recv()
        data = buffer.read()
        if len(data) < 5 or data[:3] != b'\x05\x01\x00':
            return False, protocol, '', 0, None
    else:
        is_socks5s, protocol = True, "socks5s"
        data = data[3:]

    if data[3] == 1:
        host, port = socket.inet_ntoa(data[4:8]), struct.unpack('>H', data[8:10])[0]
        if not check_allow_deny_domain(allow_host_names, deny_host_names, (host, port)):
            if not is_socks5s:
                await conn.send(b"".join([b'\x05\x02\x00\x01', socket.inet_aton('0.0.0.0'), struct.pack(">H", 0)]))
            return False, protocol, host, port, data[10:]
        if not is_socks5s:
            await conn.send(b"".join([b'\x05\x00\x00\x01', socket.inet_aton('0.0.0.0'), struct.pack(">H", 0)]))
        return True, protocol, host, port, data[10:]

    if data[3] == 4:
        host, port = socket.inet_ntop(socket.AF_INET6, data[4:20]), struct.unpack('>H', data[20:22])[0]
        if not check_allow_deny_domain(allow_host_names, deny_host_names, (host, port)):
            if not is_socks5s:
                await conn.send(b"".join([b'\x05\x02\x00\x01', socket.inet_aton('0.0.0.0'), struct.pack(">H", 0)]))
            return False, protocol, host, port, data[22:]
        if not is_socks5s:
            await conn.send(b"".join([b'\x05\x00\x00\x01', socket.inet_aton('0.0.0.0'), struct.pack(">H", 0)]))
        return True, protocol, host, port, data[22:]

    elif data[3] == 3:
        host_len = data[4] + 5
        host, port = data[5: host_len].decode("utf-8"), struct.unpack('>H', data[host_len: host_len+2])[0]
        if not check_allow_deny_domain(allow_host_names, deny_host_names, (host, port)):
            if not is_socks5s:
                await conn.send(b"".join([b'\x05\x02\x00\x01', socket.inet_aton('0.0.0.0'), struct.pack(">H", 0)]))
            return False, protocol, host, port, data[host_len + 2:]
        if not is_socks5s:
            await conn.send(b"".join([b'\x05\x00\x00\x01', socket.inet_aton('0.0.0.0'), struct.pack(">H", 0)]))
        return True, protocol, host, port, data[host_len + 2:]
    return False, protocol, '', 0, None

async def socks4_protocol_parse(conn, buffer, allow_host_names=None, deny_host_names=None):
    data = buffer.read()
    if len(data) < 9 or data[0] != 4 or data[1] != 1:
        return False, "socks4", '', 0, None
    host, port = socket.inet_ntoa(data[4:8]), struct.unpack('>H', data[2:4])[0]
    data = data[8:]
    try:
        index = data.index(b'\x00')
        data = data[index + 1:]
        index = data.index(b'\x00')
        host = data[:index].decode("utf-8")
        if not check_allow_deny_domain(allow_host_names, deny_host_names, (host, port)):
            await conn.send(b'\x00\x5b\x00\x00\x00\x00\x00\x00')
            return False, "socks4a", host, port, None
        await conn.send(b'\x00\x5a\x00\x00\x00\x00\x00\x00')
        return True, "socks4a", host, port, None
    except ValueError:
        if host.startswith("0.0.0."):
            await conn.send(b'\x00\x5b\x00\x00\x00\x00\x00\x00')
            return False, "socks4", "", 0, None
        if not check_allow_deny_domain(allow_host_names, deny_host_names, (host, port)):
            await conn.send(b'\x00\x5b\x00\x00\x00\x00\x00\x00')
            return False, "socks4", host, port, None
        await conn.send(b'\x00\x5a\x00\x00\x00\x00\x00\x00')
        return True, "socks4", host, port, None

async def tcp_proxy(conns, server, conn, status, allow_host_names=None, deny_host_names=None):
    start_time = time.time()
    host, port, protocol = '', 0, ''
    conn.write, pconn = warp_write(conn, status, "recv_len"), None
    try:
        conn.enable_nodelay()
        buffer = await conn.recv()
        if buffer[0] == 5:
            is_allowed, protocol, host, port, data = await socks5_protocol_parse(conn, buffer, allow_host_names, deny_host_names)
        elif buffer[0] == 4:
            is_allowed, protocol, host, port, data = await socks4_protocol_parse(conn, buffer, allow_host_names, deny_host_names)
        else:
            protocol = 'http'
            is_allowed, host, port, data = await http_protocol_parse(conn, buffer, allow_host_names, deny_host_names)
        if not host or not port:
            logging.info("connected %s:%s -> %s:%s empty address",
                         conn.address[0], conn.address[1], server.address[0], server.address[1])
            return
        if not is_allowed:
            logging.info("connected %s %s:%s -> %s:%s -> %s:%d not allow", protocol,
                         conn.address[0], conn.address[1], server.address[0], server.address[1], host, port)
            return

        logging.info("connected %s %s:%s -> %s:%s -> %s:%d", protocol,
                     conn.address[0], conn.address[1], server.address[0], server.address[1], host, port)
        pconn = create_socket((host, port))
        pconn.write = warp_write(pconn, status, "send_len")
        await pconn.connectof((host, port))
        if data:
            await pconn.send(data)
        await pconn.linkof(conn)
    except sevent.errors.SocketClosed:
        pass
    except Exception as e:
        logging.info("error %s %s:%s -> %s:%s -> %s:%d %s %.2fms\r%s", protocol,
                     conn.address[0], conn.address[1], server.address[0], server.address[1], host, port, e,
                     (time.time() - start_time) * 1000, traceback.format_exc())
        return
    finally:
        conn.close()
        if pconn: pconn.close()
        conns.pop(id(conn), None)

    logging.info("closed %s %s:%s -> %s:%s -> %s:%d %s %s %.2fms", protocol,
                 conn.address[0], conn.address[1], server.address[0], server.address[1], host, port, format_data_len(status["send_len"]),
                 format_data_len(status["recv_len"]), (time.time() - start_time) * 1000)

async def check_timeout(conns, timeout):
    def run_check():
        while True:
            try:
                now = time.time()
                for conn_id, (conn, status) in list(conns.items()):
                    if status['check_recv_len'] != status['recv_len'] or status['check_send_len'] != status['send_len']:
                        status["check_recv_len"] = status["recv_len"]
                        status["check_send_len"] = status["send_len"]
                        status['last_time'] = now
                        continue

                    if now - status['last_time'] >= timeout:
                        sevent.current().add_async_safe(conn.close)
                        conns.pop(conn_id, None)
            finally:
                time.sleep(min(float(timeout) / 2.0, 30))

    if timeout > 0:
        check_thread = threading.Thread(target=run_check)
        check_thread.daemon = True
        check_thread.start()
    await sevent.Future()

async def tcp_accept(server, timeout, allow_host_names=None, deny_host_names=None):
    conns = {}
    sevent.current().call_async(check_timeout, conns, timeout)
    while True:
        conn = await server.accept()
        status = {"recv_len": 0, "send_len": 0, "last_time": time.time(), "check_recv_len": 0, "check_send_len": 0}
        sevent.current().call_async(tcp_proxy, conns, server, conn, status, allow_host_names, deny_host_names)
        conns[id(conn)] = (conn, status)

def main(argv):
    parser = argparse.ArgumentParser(description='simple http and socks5 proxy server')
    parser.add_argument('-b', dest='bind', default="0.0.0.0", help='local bind host (default: 0.0.0.0)')
    parser.add_argument('-p', dest='port', default=8088, type=int, help='local bind port (default: 8088)')
    parser.add_argument('-t', dest='timeout', default=7200, type=int, help='no read/write timeout (default: 7200)')
    parser.add_argument('-a', dest='allow_hosts', default=[], action="append", type=str,help='allow forward host name')
    parser.add_argument('-d', dest='deny_hosts', default=[], action="append", type=str, help='deny forward host name')
    parser.add_argument('-A', dest='allow_filename', default="", type=str, help='allow forward host name config filename')
    parser.add_argument('-D', dest='deny_filename', default="", type=str, help='deny forward host name config filename')
    args = parser.parse_args(args=argv)
    config_signal()

    allow_host_names, deny_host_names = [], []
    if args.allow_hosts:
        for allow_host in args.allow_hosts:
            allow_host_names.extend([parse_allow_deny_host(h) for h in allow_host.split(",") if h.strip()])
    if args.deny_hosts:
        for deny_host in args.deny_hosts:
            deny_host_names.extend([parse_allow_deny_host(h) for h in deny_host.split(",") if h.strip()])
    if args.allow_filename:
        allow_host_names.extend(parse_allow_deny_host_filename(args.allow_filename))
    if args.deny_filename:
        deny_host_names.extend(parse_allow_deny_host_filename(args.deny_filename))

    logging.info("listen server at %s:%d", args.bind, args.port)
    server = create_server((args.bind, args.port))
    sevent.current().call_async(tcp_accept, server, args.timeout, allow_host_names or None, deny_host_names or None)

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)1.1s %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S', filemode='a+')
    try:
        main(sys.argv[1:])
        sevent.instance().start()
    except KeyboardInterrupt:
        exit(0)


# -*- coding: utf-8 -*-
# 2021/2/1
# create by: snower

import shlex
import sys
import multiprocessing
import logging
import sevent
from . import tcp_forward
from . import simple_proxy
from . import tcp2proxy
from . import proxy2proxy
from . import redirect2proxy
from . import tcp_reverse
from . import arproxy
from . import proxy_tunnel

HEPERS = {
    "tcp_forward": tcp_forward,
    "simple_proxy": simple_proxy,
    "tcp2proxy": tcp2proxy,
    "proxy2proxy": proxy2proxy,
    "redirect2proxy": redirect2proxy,
    "tcp_reverse": tcp_reverse,
    "arproxy": arproxy,
    "proxy_tunnel": proxy_tunnel,
}

def show_help_message():
    print('usage: -m [HELPER_NAME] [ARGS]\r\n')
    print('simple sevent helpers \r\n')
    print("can use helpers:\r\n\r\n" + '\r\n'.join(["sevent.helpers." + name for name in HEPERS]))
    print('\r\n\r\n' + '*' * 64 + '\r\n')

    for name, helper in HEPERS.items():
        sys.argv[0] = "-m sevent.helpers." + name
        p = multiprocessing.Process(target=helper.main, args=(["-h"],))
        p.start()
        p.join()
        print('\r\n\r\n' + '*' * 64 + '\r\n')

if __name__ == "__main__":
    if "-h" in sys.argv:
        show_help_message()
        exit(0)

    args_helpers = []
    if len(sys.argv) >= 2 and sys.argv[1] == "-f":
        if len(sys.argv) >= 3:
            with open(sys.argv[2], "r", encoding="utf-8") as fp:
                for line in fp:
                    args = shlex.split(line)
                    if not args:
                        continue
                    if args[0] and args[0][0] == "@" and args[0][1:] in HEPERS:
                        args_helpers.append((args[0][1:], HEPERS[args[0][1:]], args[1:]))
    else:
        for arg in sys.argv[1:]:
            if arg and arg[0] == "@" and arg[1:] in HEPERS:
                args_helpers.append((arg[1:], HEPERS[arg[1:]], []))
            elif not args_helpers:
                continue
            else:
                args_helpers[-1][2].append(arg)

    if not args_helpers:
        show_help_message()
        exit(0)

    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)1.1s %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S', filemode='a+')
    for name, helper, argv in args_helpers:
        logging.info("start helper %s by %s", name, argv)
        helper.main(argv)
    try:
        sevent.instance().start()
    except KeyboardInterrupt:
        exit(0)


# -*- coding: utf-8 -*-
# 2025/2/8
# create by: snower

import ssl
import time

from ..buffer import Buffer, BaseBuffer, RECV_BUFFER_SIZE
from ..errors import SSLConnectError, SSLSocketError, SocketClosed, ConnectTimeout
from ..tcp import WarpSocket, WarpServer, STATE_INITIALIZED, STATE_CONNECTING, STATE_CLOSED


class SSLSocket(WarpSocket):
    _default_context = None

    @classmethod
    def load_default_context(cls):
        if cls._default_context is None:
            cls._default_context = ssl.create_default_context()
        return cls._default_context

    def __init__(self, context=None, server_side=False, server_hostname=None, session=None, *args, **kwargs):
        WarpSocket.__init__(self, *args, **kwargs)

        self._context = context or self.load_default_context()
        self._server_side = server_side
        self._handshaked = False
        self._shutdowned = None
        self._incoming = ssl.MemoryBIO()
        self._outgoing = ssl.MemoryBIO()
        self._ssl_bio = self._context.wrap_bio(self._incoming, self._outgoing, server_side, server_hostname, session)
        self._connect_timeout = 5
        self._connect_timestamp = 0
        self._handshake_callback = None
        self._handshake_timeout_handler = None
        self._shutdown_timeout_handler = None

    @property
    def context(self):
        return self._context

    @property
    def sslobj(self):
        return self._ssl_bio

    @property
    def session(self):
        if self._ssl_bio is not None:
            return self._ssl_bio.session
        return None

    @property
    def server_side(self):
        return self._server_side

    @property
    def server_hostname(self):
        if self._ssl_bio is not None:
            return self._ssl_bio.server_hostname
        return None

    def connect(self, address, timeout=5):
        if self._state != STATE_INITIALIZED:
            if self._state == STATE_CLOSED:
                raise SocketClosed()
            return
        self._connect_timeout = timeout
        self._connect_timestamp = time.time()
        WarpSocket.connect(self, address, timeout)

    def close(self):
        if self._state == STATE_CLOSED or self._shutdowned is False:
            return
        if not self._handshaked or self._shutdowned is True:
            self._shutdowned = True
            WarpSocket.close(self)
            return
        def on_timeout_cb():
            self._shutdown_timeout_handler = None
            if self._shutdowned is True:
                return
            self._shutdowned = True
            self._error(ConnectTimeout("ssl shutdown time out %s" % str(self.address)))
        self._shutdown_timeout_handler = self._loop.add_timeout(30, on_timeout_cb)
        self._shutdowned = False
        self.do_shutdown()

    def _do_connect(self, socket):
        if self._state not in (STATE_INITIALIZED, STATE_CONNECTING):
            self._shutdowned = True
            WarpSocket.close(self)
            return
        if self._handshaked:
            WarpSocket._do_connect(self, socket)
            return
        timeout = max(1.0, self._connect_timeout - (time.time() - self._connect_timestamp))
        self.start_handshake(timeout, lambda _: self._loop.add_async(WarpSocket._do_connect, self, self._socket))

    def _do_close(self, socket):
        if self._ssl_bio is None or self._context is None:
            return
        self._incoming = None
        self._outgoing = None
        self._ssl_bio = None
        self._context = None
        self._shutdowned = True
        if self._handshake_timeout_handler:
            self._loop.cancel_timeout(self._handshake_timeout_handler)
            self._handshake_timeout_handler = None
        if self._shutdown_timeout_handler:
            self._loop.cancel_timeout(self._shutdown_timeout_handler)
            self._shutdown_timeout_handler = None
        WarpSocket._do_close(self, socket)

    def start_handshake(self, timeout, handshake_callback):
        self._handshake_callback = handshake_callback
        def on_timeout_cb():
            self._handshake_timeout_handler = None
            if self._handshaked:
                return
            self._error(ConnectTimeout("ssl handshake time out %s" % str(self.address)))
        self._handshake_timeout_handler = self._loop.add_timeout(max(1, timeout), on_timeout_cb)
        self.do_handshake()

    def read(self, data):
        if self._state == STATE_CLOSED:
            return
        if data.__class__ is Buffer:
            data = data.read()
        n = self._incoming.write(data)
        if n <= 0:
            if not data:
                return
            self._shutdowned = True
            self._loop.add_async(self._error, SSLSocketError("incoming write return zero"))
            return

        if not self._handshaked:
            while True:
                if self.do_handshake():
                    if n >= len(data):
                        return
                    nn = self._incoming.write(data[n:])
                    if nn <= 0:
                        self._shutdowned = True
                        self._loop.add_async(self._error, SSLSocketError("incoming write return zero"))
                        return
                    n += nn
                else:
                    break

        try:
            last_data_len = self._rbuffers._len
            while True:
                while True:
                    try:
                        chunk = self._ssl_bio.read(RECV_BUFFER_SIZE)
                        if not chunk:
                            break
                        BaseBuffer.write(self._rbuffers, chunk)
                    except ssl.SSLWantReadError:
                        if self._outgoing.pending:
                            self.flush()
                        break
                    except ssl.SSLWantWriteError:
                        if self._outgoing.pending:
                            self.flush()
                    except (ssl.SSLZeroReturnError, ssl.SSLEOFError):
                        self._shutdowned = True
                        self._loop.add_async(self.close)
                        return
                if n >= len(data):
                    break
                nn = self._incoming.write(data[n:])
                if nn <= 0:
                    self._shutdowned = True
                    self._loop.add_async(self._error, SSLSocketError("incoming write return zero"))
                    return
                n += nn
            if last_data_len < self._rbuffers._len:
                if self._rbuffers._len > self._rbuffers._drain_size and not self._rbuffers._full:
                    self._rbuffers.do_drain()
                self.emit_data(self, self._rbuffers)
        except Exception as e:
            self._shutdowned = True
            self._loop.add_async(self._error, SSLSocketError(str(e)))

    def write(self, data):
        if self._state == STATE_CLOSED:
            raise SocketClosed()
        try:
            if not self._handshaked:
                if data.__class__ is Buffer:
                    BaseBuffer.extend(self._wbuffers, data)
                    if data._full and data._len < data._regain_size:
                        data.do_regain()
                else:
                    BaseBuffer.write(self._wbuffers, data)
                if self._wbuffers._len > self._wbuffers._drain_size and not self._wbuffers._full:
                    self._wbuffers.do_drain()
                return False

            if data.__class__ is Buffer:
                data = data.read()
            try:
                n = self._ssl_bio.write(data)
                if n <= 0 and data:
                    if not self._outgoing.pending:
                        raise SSLSocketError("outgoing write return zero")
                    self.flush()
            except (ssl.SSLWantWriteError, ssl.SSLWantReadError):
                if self._outgoing.pending:
                    self.flush()
                n = 0

            while n < len(data):
                try:
                    nn = self._ssl_bio.write(data[n:])
                    if nn <= 0:
                        if not self._outgoing.pending:
                            raise SSLSocketError("outgoing write return zero")
                        self.flush()
                        continue
                    n += nn
                except (ssl.SSLWantWriteError, ssl.SSLWantReadError):
                    if self._outgoing.pending:
                        self.flush()
        except (ssl.SSLZeroReturnError, ssl.SSLEOFError) as e:
            self._shutdowned = True
            self._loop.add_async(self.close)
            raise e
        except Exception as e:
            self._shutdowned = True
            self._loop.add_async(self._error, SSLSocketError(str(e)) if not isinstance(e, SSLSocketError) else e)
            raise e
        if self._outgoing.pending:
            return self.flush()
        return True

    def do_handshake(self):
        while True:
            try:
                self._ssl_bio.do_handshake()
                if self._outgoing.pending:
                    self.flush()
                self._handshaked = True
                if self._wbuffers:
                    self.write(self._wbuffers)
                if self._handshake_timeout_handler:
                    self._loop.cancel_timeout(self._handshake_timeout_handler)
                    self._handshake_timeout_handler = None
                handshake_callback, self._handshake_callback = self._handshake_callback, None
                if handshake_callback is not None:
                    handshake_callback(self)
                return False
            except ssl.SSLWantReadError:
                if self._outgoing.pending:
                    self.flush()
                break
            except ssl.SSLWantWriteError:
                if self._outgoing.pending:
                    self.flush()
            except Exception as e:
                self._shutdowned = True
                self._loop.add_async(self._error, SSLConnectError(self.address, e, "ssl handshake error %s %s" % (str(self.address), e)))
                break
        return True

    def do_shutdown(self):
        if self._shutdowned is True:
            return
        while True:
            try:
                if self._handshaked:
                    self._ssl_bio.unwrap()
                    if self._outgoing.pending:
                        self.flush()
                self._shutdowned = True
                WarpSocket.close(self)
                if self._shutdown_timeout_handler:
                    self._loop.cancel_timeout(self._shutdown_timeout_handler)
                    self._shutdown_timeout_handler = None
                break
            except ssl.SSLWantReadError:
                if self._outgoing.pending:
                    self.flush()
                break
            except ssl.SSLWantWriteError:
                if self._outgoing.pending:
                    self.flush()
            except Exception as e:
                self._shutdowned = True
                self._loop.add_async(self._error, SSLSocketError(str(e)))
                break

    def flush(self):
        data = self._outgoing.read()
        if not data:
            return True
        last_state = WarpSocket.write(self, data)
        data = self._outgoing.read()
        if not data:
            return last_state
        last_state = WarpSocket.write(self, data)
        while True:
            data = self._outgoing.read()
            if not data:
                return last_state
            last_state = WarpSocket.write(self, data)


class SSLServer(WarpServer):
    @classmethod
    def create_server_context(cls):
        return ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)

    def __init__(self, context, *args, **kwargs):
        WarpServer.__init__(self, *args, **kwargs)

        self._context = context

    @property
    def context(self):
        return self._context

    def handshake(self, socket):
        max_buffer_size = socket._max_buffer_size if hasattr(socket, "_max_buffer_size") else None
        socket = SSLSocket(context=self._context, server_side=True, socket=socket, loop=self._loop, max_buffer_size=max_buffer_size)
        socket.start_handshake(30, lambda _: self._loop.add_async(self.emit_connection, self, socket))


# -*- coding: utf-8 -*-
# 2021/2/1
# create by: snower

import greenlet
from collections import deque
from ..errors import ChainClosed
from ..loop import current

class Chain(object):
    def __init__(self, size=1):
        self._size = size
        self._queue = deque()
        self._queue_size = 0
        self._closed = False
        self._send_waiters = deque()
        self._recv_waiters = deque()

    async def send(self, value):
        if self._closed:
            raise ChainClosed()

        if self._recv_waiters:
            current().add_async(self._recv_waiters.popleft().switch, value)
            return

        self._queue.append(value)
        self._queue_size += 1
        if self._queue_size <= self._size:
            return

        child_gr = greenlet.getcurrent()
        main = child_gr.parent
        assert main is not None, "must be running in async func"
        self._send_waiters.append(child_gr)
        main.switch()

    async def recv(self):
        if self._closed:
            raise ChainClosed()

        if self._queue_size > 0:
            if self._send_waiters:
                current().add_async(self._send_waiters.popleft().switch)
            self._queue_size -= 1
            return self._queue.popleft()

        child_gr = greenlet.getcurrent()
        main = child_gr.parent
        assert main is not None, "must be running in async func"
        self._recv_waiters.append(child_gr)
        return main.switch()

    def close(self):
        if self._closed:
            return

        self._closed = True
        while self._send_waiters:
            current().add_async(self._send_waiters.popleft().throw, ChainClosed())
        while self._recv_waiters:
            current().add_async(self._recv_waiters.popleft().throw, ChainClosed())
        self._queue_size = 0
        self._queue.clear()

    async def closeof(self):
        self.close()

    def __del__(self):
        self.close()


# -*- coding: utf-8 -*-
# 2022/1/21
# create by: snower

from .utils import is_py3, get_logger
from .event import EventEmitter, null_emit_callback
from .loop import instance
from .buffer import Buffer, BaseBuffer, RECV_BUFFER_SIZE
from .errors import SocketClosed

STATE_INITIALIZED = 0x01
STATE_CONNECTING = 0x02
STATE_STREAMING = 0x04
STATE_LISTENING = 0x08
STATE_CLOSING = 0x10
STATE_CLOSED = 0x20

class PipeSocket(EventEmitter):
    MAX_BUFFER_SIZE = None
    RECV_BUFFER_SIZE = RECV_BUFFER_SIZE

    @classmethod
    def config(cls, max_buffer_size=None, recv_buffer_size=RECV_BUFFER_SIZE, **kwargs):
        cls.MAX_BUFFER_SIZE = max_buffer_size
        cls.RECV_BUFFER_SIZE = recv_buffer_size

    def __init__(self, loop=None, socket=None, address=None, max_buffer_size=None):
        EventEmitter.__init__(self)
        self._loop = loop or instance()
        self._socket = socket
        self._address = address
        self._max_buffer_size = max_buffer_size or self.MAX_BUFFER_SIZE
        self._rbuffers = Buffer(max_buffer_size=self._max_buffer_size)
        self._wbuffers = Buffer(max_buffer_size=self._max_buffer_size)
        self._state = STATE_STREAMING if socket else STATE_INITIALIZED
        self._has_drain_event = False
        self.ignore_write_closed_error = False

        if self._state == STATE_STREAMING:
            self._rbuffers.on("drain", lambda _: self.drain())
            self._rbuffers.on("regain", lambda _: self.regain())

    @property
    def address(self):
        return self._address

    @property
    def socket(self):
        return self

    @property
    def buffer(self):
        return self._rbuffers, self._wbuffers

    def __del__(self):
        self.close()

    def on(self, event_name, callback):
        EventEmitter.on(self, event_name, callback)

        if event_name == "drain":
            self._has_drain_event = True

    def once(self, event_name, callback):
        EventEmitter.once(self, event_name, callback)

        if event_name == "drain":
            self._has_drain_event = True

    def off(self, event_name, callback):
        EventEmitter.off(self, event_name, callback)

        if not self._events[event_name] and not self._events_once[event_name]:
            if event_name == "drain":
                self._has_drain_event = False

    def noce(self, event_name, callback):
        EventEmitter.noce(self, event_name, callback)

        if not self._events[event_name] and not self._events_once[event_name]:
            if event_name == "drain":
                self._has_drain_event = False

    def remove_listener(self, event_name, callback):
        EventEmitter.remove_listener(self, event_name, callback)

        if not self._events[event_name] and not self._events_once[event_name]:
            if event_name == "drain":
                self._has_drain_event = False

    def on_connect(self, callback):
        self.on("connect", callback)

    def on_data(self, callback):
        self.on("data", callback)

    def on_end(self, callback):
        self.on("end", callback)

    def on_close(self, callback):
        self.on("close", callback)

    def on_error(self, callback):
        self.on("error", callback)

    def on_drain(self, callback):
        self.on("drain", callback)

    def off_connect(self, callback):
        self.on("connect", callback)

    def off_data(self, callback):
        self.on("data", callback)

    def off_end(self, callback):
        self.on("end", callback)

    def off_close(self, callback):
        self.on("close", callback)

    def off_error(self, callback):
        self.on("error", callback)

    def off_drain(self, callback):
        self.on("drain", callback)

    def once_connect(self, callback):
        self.once("connect", callback)

    def once_data(self, callback):
        self.once("data", callback)

    def once_end(self, callback):
        self.once("end", callback)

    def once_close(self, callback):
        self.once("close", callback)

    def once_error(self, callback):
        self.once("error", callback)

    def once_drain(self, callback):
        self.once("drain", callback)

    def noce_connect(self, callback):
        self.once("connect", callback)

    def noce_data(self, callback):
        self.once("data", callback)

    def noce_end(self, callback):
        self.once("end", callback)

    def noce_close(self, callback):
        self.once("close", callback)

    def noce_error(self, callback):
        self.once("error", callback)

    def noce_drain(self, callback):
        self.once("drain", callback)

    def enable_fast_open(self):
        pass

    @property
    def is_enable_fast_open(self):
        return False

    def enable_nodelay(self):
        pass

    @property
    def is_enable_nodelay(self):
        return True

    def end(self):
        if self._state not in (STATE_INITIALIZED, STATE_CONNECTING, STATE_STREAMING):
            return

        if self._state in (STATE_INITIALIZED, STATE_CONNECTING):
            self._loop.add_async(self.close)
        else:
            if self._wbuffers:
                self._state = STATE_CLOSING
            else:
                self._loop.add_async(self.close)

    def close(self):
        if self._state == STATE_CLOSED:
            return

        self._state = STATE_CLOSED
        def on_close():
            try:
                self.emit_close(self)
            except Exception as e:
                get_logger().exception("tcp emit close error:%s", e)
            self.remove_all_listeners()
            self._rbuffers.close()
            self._wbuffers.close()
            self._rbuffers = None
            self._wbuffers = None
            self._socket = None
        self._loop.add_async(on_close)
        if self._socket:
            self._loop.add_async(self._socket.end)

    def _error(self, error):
        self._loop.add_async(self.emit_error, self, error)
        self._loop.add_async(self.close)
        if self.emit_error == null_emit_callback:
            get_logger().error("Pipe %s socket %s error: %s", self, self.socket, error)

    def connect(self, address, timeout=5):
        if self._state != STATE_INITIALIZED:
            if self._state == STATE_CLOSED:
                raise SocketClosed()
            return
        if isinstance(address, (tuple, list)):
            address = "pipe#%s" % (address[1] if len(address) >= 2 else address[-1])
        elif not isinstance(address, str):
            address = "pipe#%s" % address
        self._socket = PipeServer._bind_servers[address]._accept_cb(self, address)
        self._address = (address, id(self))
        self._state = STATE_STREAMING
        self._rbuffers.on("drain", lambda _: self.drain())
        self._rbuffers.on("regain", lambda _: self.regain())
        self._loop.add_async(self.emit_connect, self)

    def drain(self):
        if self._state in (STATE_STREAMING, STATE_CLOSING):
            self._socket._wbuffers.do_drain()

    def regain(self):
        if self._state in (STATE_STREAMING, STATE_CLOSING):
            self._socket._wbuffers.do_regain()

    def write(self, data):
        if self._state != STATE_STREAMING:
            if self.ignore_write_closed_error:
                return False
            raise SocketClosed()

        socket = self._socket
        if data.__class__ is Buffer:
            BaseBuffer.extend(socket._rbuffers, data)
        else:
            BaseBuffer.write(socket._rbuffers, data)
        if socket._rbuffers._len > socket._rbuffers._drain_size and not socket._rbuffers._full:
            socket._rbuffers.do_drain()
        socket.emit_data(socket, socket._rbuffers)
        if self._has_drain_event and not socket._rbuffers:
            self._loop.add_async(self.emit_drain, self)
        return True

    def link(self, socket):
        if self._state not in (STATE_STREAMING, STATE_CONNECTING):
            raise SocketClosed()
        if socket._state not in (STATE_STREAMING, STATE_CONNECTING):
            raise SocketClosed()

        self.ignore_write_closed_error = True
        socket.ignore_write_closed_error = True
        rbuffer, wbuffer = socket.buffer
        if self._state != STATE_STREAMING:
            if self.is_enable_fast_open and rbuffer:
                self.write(rbuffer)

            def on_connect(s):
                self._wbuffers.link(rbuffer)
                if rbuffer:
                    self.write(rbuffer)
                socket.on_data(lambda s, data: self.write(data))

            self.on_connect(on_connect)

        else:
            self._wbuffers.link(rbuffer)
            if rbuffer:
                self.write(rbuffer)
            socket.on_data(lambda s, data: self.write(data))

        if socket._state != STATE_STREAMING:
            if socket.is_enable_fast_open and self._rbuffers:
                socket.write(self._rbuffers)

            def on_pconnect(s):
                wbuffer.link(self._rbuffers)
                if self._rbuffers:
                    socket.write(self._rbuffers)
                self.on_data(lambda s, data: socket.write(data))

            socket.on_connect(on_pconnect)
        else:
            wbuffer.link(self._rbuffers)
            if self._rbuffers:
                socket.write(self._rbuffers)
            self.on_data(lambda s, data: socket.write(data))

        self.on_close(lambda s: socket.end())
        socket.on_close(lambda s: self.end())


class PipeServer(EventEmitter):
    _bind_servers = {}

    def __init__(self, loop=None):
        EventEmitter.__init__(self)
        self._loop = loop or instance()
        self._state = STATE_INITIALIZED
        self._address = None

    def __del__(self):
        self.close()

    def on_listen(self, callback):
        self.on("listen", callback)

    def on_connection(self, callback):
        self.on("connection", callback)

    def on_close(self, callback):
        self.on("close", callback)

    def on_error(self, callback):
        self.on("error", callback)

    def once_listen(self, callback):
        self.on("listen", callback)

    def once_connection(self, callback):
        self.on("connection", callback)

    def once_close(self, callback):
        self.on("close", callback)

    def once_error(self, callback):
        self.on("error", callback)

    def enable_fast_open(self):
        pass

    def enable_reuseaddr(self):
        pass

    def enable_nodelay(self):
        pass

    @property
    def is_enable_fast_open(self):
        return False

    @property
    def is_reuseaddr(self):
        return False

    @property
    def is_enable_nodelay(self):
        return True

    def listen(self, address, backlog=128):
        if self._state != STATE_INITIALIZED:
            if self._state == STATE_CLOSED:
                raise SocketClosed()
            return
        if isinstance(address, (tuple, list)):
            address = "pipe#%s" % (address[1] if len(address) >= 2 else address[-1])
        elif not isinstance(address, str):
            address = "pipe#%s" % address
        PipeServer._bind_servers[address] = self
        self._address = address
        self._state = STATE_LISTENING

    def _accept_cb(self, socket, address):
        if self._state != STATE_LISTENING:
            return
        socket = PipeSocket(socket=socket, address=(address, id(socket)))
        self._loop.add_async(self.emit_connection, self, socket)
        return socket

    def _error(self, error):
        self._loop.add_async(self.emit_error, self, error)
        self.close()
        get_logger().error("server error: %s", error)

    def close(self):
        if self._state in (STATE_INITIALIZED, STATE_LISTENING):
            if self._address and self._address in PipeServer._bind_servers:
                PipeServer._bind_servers.pop(self._address)
            self._state = STATE_CLOSED

            def on_close():
                try:
                    self.emit_close(self)
                except Exception as e:
                    get_logger().exception("tcp server emit close error:%s", e)
                self.remove_all_listeners()
            self._loop.add_async(on_close)


if is_py3:
    from .coroutines.pipe import warp_coroutine
    PipeSocket, PipeServer = warp_coroutine(PipeSocket, PipeServer)


# -*- coding: utf-8 -*-
# 2020/5/8
# create by: snower

import greenlet
from ..errors import SocketClosed

STATE_INITIALIZED = 0x01
STATE_CONNECTING = 0x02
STATE_STREAMING = 0x04
STATE_LISTENING = 0x08
STATE_CLOSING = 0x10
STATE_CLOSED = 0x20


def warp_coroutine(BaseSocket, BaseServer):
    class PipeSocket(BaseSocket):
        _connect_greenlet = None
        _send_greenlet = None
        _recv_greenlet = None
        _close_error_registed = False
        _recv_size = 0

        def _on_close_handle(self, socket):
            if self._connect_greenlet is not None:
                self.off("connect", self._on_connect_handle)
                child_gr, self._connect_greenlet = self._connect_greenlet, None
                child_gr.throw(SocketClosed())
            if self._send_greenlet is not None:
                self.off("drain", self._on_send_handle)
                child_gr, self._send_greenlet = self._send_greenlet, None
                child_gr.throw(SocketClosed())
            if self._recv_greenlet is not None:
                self.off("data", self._on_recv_handle)
                child_gr, self._recv_greenlet = self._recv_greenlet, None
                child_gr.throw(SocketClosed())

        def _on_error_handle(self, socket, e):
            if self._connect_greenlet is not None:
                self.off("connect", self._on_connect_handle)
                child_gr, self._connect_greenlet = self._connect_greenlet, None
                child_gr.throw(e)
            if self._send_greenlet is not None:
                self.off("drain", self._on_send_handle)
                child_gr, self._send_greenlet = self._send_greenlet, None
                child_gr.throw(e)
            if self._recv_greenlet is not None:
                self.off("data", self._on_recv_handle)
                child_gr, self._recv_greenlet = self._recv_greenlet, None
                child_gr.throw(e)

        def _on_connect_handle(self, socket):
            if self._connect_greenlet is None:
                return
            self.off("connect", self._on_connect_handle)
            if self._close_error_registed:
                self.off("close", self._on_close_handle)
                self.off("error", self._on_error_handle)
                self._close_error_registed = False
            child_gr, self._connect_greenlet = self._connect_greenlet, None
            return child_gr.switch()

        def _on_send_handle(self, socket):
            if self._send_greenlet is None:
                return
            self.off("drain", self._on_send_handle)
            child_gr, self._send_greenlet = self._send_greenlet, None
            return child_gr.switch()

        def _on_recv_handle(self, socket, buffer):
            if self._recv_greenlet is None:
                return
            if len(buffer) < self._recv_size:
                return
            self.off("data", self._on_recv_handle)
            child_gr, self._recv_greenlet = self._recv_greenlet, None
            return child_gr.switch(buffer)

        async def connectof(self, address, timeout=5):
            assert self._connect_greenlet is None, "already connecting"
            if self._state != STATE_INITIALIZED:
                if self._state == STATE_CLOSED:
                    raise SocketClosed()
                return
            if not self._close_error_registed:
                self.on("close", self._on_close_handle)
                self.on("error", self._on_error_handle)
                self._close_error_registed = True

            self._connect_greenlet = greenlet.getcurrent()
            main = self._connect_greenlet.parent
            assert main is not None, "must be running in async func"

            self.on("connect", self._on_connect_handle)
            self.connect(address, timeout)
            return main.switch()

        async def send(self, data):
            assert self._send_greenlet is None, "already sending"
            if self._state == STATE_CLOSED:
                raise SocketClosed()
            if not self._close_error_registed:
                self.on("close", self._on_close_handle)
                self.on("error", self._on_error_handle)
                self._close_error_registed = True

            if self.write(data):
                return

            self._send_greenlet = greenlet.getcurrent()
            main = self._send_greenlet.parent
            assert main is not None, "must be running in async func"
            self.on("drain", self._on_send_handle)
            return main.switch()

        async def recv(self, size=0):
            assert self._recv_greenlet is None, "already recving"
            if self._state == STATE_CLOSED:
                raise SocketClosed()
            if not self._close_error_registed:
                self.on("close", self._on_close_handle)
                self.on("error", self._on_error_handle)
                self._close_error_registed = True

            if self._rbuffers and len(self._rbuffers) >= size:
                return self._rbuffers

            self._recv_greenlet = greenlet.getcurrent()
            main = self._recv_greenlet.parent
            assert main is not None, "must be running in async func"

            self.on("data", self._on_recv_handle)
            self._recv_size = size
            if self._rbuffers._drain_size < size:
                drain_size, self._rbuffers._drain_size = self._rbuffers._drain_size, size
                try:
                    if self._rbuffers._full:
                        self._rbuffers.do_regain()
                    return main.switch()
                finally:
                    self._rbuffers._drain_size = drain_size
                    if self._rbuffers._len > self._rbuffers._drain_size and not self._rbuffers._full:
                        self._rbuffers.do_drain()
                    self._recv_size = 0
            try:
                if self._rbuffers._full:
                    self._rbuffers.do_regain()
                return main.switch()
            finally:
                self._recv_size = 0

        async def closeof(self):
            if self._state == STATE_CLOSED:
                return

            child_gr = greenlet.getcurrent()
            main = child_gr.parent
            assert main is not None, "must be running in async func"

            self.on("close", lambda socket: child_gr.switch())
            self.end()
            return main.switch()

        async def linkof(self, socket):
            assert self._connect_greenlet is None, "already connecting"
            assert self._send_greenlet is None, "already sending"
            assert self._recv_greenlet is None, "already recving"
            if self._state not in (STATE_STREAMING, STATE_CONNECTING):
                raise SocketClosed()
            if socket._state not in (STATE_STREAMING, STATE_CONNECTING):
                raise SocketClosed()
            if self._close_error_registed:
                self.off("close", self._on_close_handle)
                self.off("error", self._on_error_handle)
                self._close_error_registed = False
            if hasattr(socket, "_close_error_registed") and socket._close_error_registed:
                socket.off("close", socket._on_close_handle)
                socket.off("error", socket._on_error_handle)
                socket._close_error_registed = False

            child_gr = greenlet.getcurrent()
            main = child_gr.parent
            assert main is not None, "must be running in async func"

            def do_closed(s):
                if self._state != STATE_CLOSED:
                    return
                if socket._state != STATE_CLOSED:
                    return
                child_gr.switch()

            self.on("close", do_closed)
            socket.on("close", do_closed)
            BaseSocket.link(self, socket)
            return main.switch()

        async def join(self):
            if self._state == STATE_CLOSED:
                return

            child_gr = greenlet.getcurrent()
            main = child_gr.parent
            assert main is not None, "must be running in async func"

            self.on("close", lambda socket: child_gr.switch())
            return main.switch()

    class PipeServer(BaseServer):
        _listen_greenlet = None
        _accept_greenlet = None
        _close_error_registed = False

        def _on_close_handle(self, socket):
            if self._listen_greenlet is not None:
                self.off("listen", self._on_listen_handle)
                child_gr, self._listen_greenlet = self._listen_greenlet, None
                child_gr.throw(SocketClosed())
            if self._accept_greenlet is not None:
                self.off("connection", self._on_accept_handle)
                child_gr, self._accept_greenlet = self._accept_greenlet, None
                child_gr.throw(SocketClosed())

        def _on_error_handle(self, socket, e):
            if self._listen_greenlet is not None:
                self.off("listen", self._on_listen_handle)
                child_gr, self._listen_greenlet = self._listen_greenlet, None
                child_gr.throw(e)
            if self._accept_greenlet is not None:
                self.off("connection", self._on_accept_handle)
                child_gr, self._accept_greenlet = self._accept_greenlet, None
                child_gr.throw(e)

        def _on_listen_handle(self, server):
            if self._listen_greenlet is None:
                return
            self.off("listen", self._on_listen_handle)
            child_gr, self._listen_greenlet = self._listen_greenlet, None
            return child_gr.switch()

        def _on_accept_handle(self, server, connection):
            if self._accept_greenlet is None:
                return
            self.off("connection", self._on_accept_handle)
            child_gr, self._accept_greenlet = self._accept_greenlet, None
            return child_gr.switch(connection)

        async def listenof(self, address, backlog=128):
            assert self._listen_greenlet is None, "already listening"
            if self._state != STATE_INITIALIZED:
                if self._state == STATE_CLOSED:
                    raise SocketClosed()
                return
            if not self._close_error_registed:
                self.on("close", self._on_close_handle)
                self.on("error", self._on_error_handle)
                self._close_error_registed = True

            self._listen_greenlet = greenlet.getcurrent()
            main = self._listen_greenlet.parent
            assert main is not None, "must be running in async func"

            self.on("listen", self._on_listen_handle)
            self.listen(address, backlog)
            return main.switch()

        async def accept(self):
            assert self._accept_greenlet is None, "already accepting"
            if self._state == STATE_CLOSED:
                raise SocketClosed()
            if not self._close_error_registed:
                self.on("close", self._on_close_handle)
                self.on("error", self._on_error_handle)
                self._close_error_registed = True

            self._accept_greenlet = greenlet.getcurrent()
            main = self._accept_greenlet.parent
            assert main is not None, "must be running in async func"

            self.on("connection", self._on_accept_handle)
            return main.switch()

        async def closeof(self):
            if self._state == STATE_CLOSED:
                return

            child_gr = greenlet.getcurrent()
            main = child_gr.parent
            assert main is not None, "must be running in async func"

            self.on("close", lambda server: child_gr.switch())
            self.close()
            return main.switch()

        async def join(self):
            if self._state == STATE_CLOSED:
                return

            child_gr = greenlet.getcurrent()
            main = child_gr.parent
            assert main is not None, "must be running in async func"

            self.on("close", lambda server: child_gr.switch())
            return main.switch()

    return PipeSocket, PipeServer



# -*- coding: utf-8 -*-
# 15/8/6
# create by: snower

import sys
import os
import time
import socket
import dnslib
from collections import defaultdict
from .loop import instance
from .event import EventEmitter
from .utils import ensure_bytes, is_py3

QTYPE_ANY = 255
QTYPE_A = 1
QTYPE_AAAA = 28
QTYPE_CNAME = 5
QTYPE_NS = 2
QCLASS_IN = 1

STATUS_OPENED = 0
STATUS_CLOSED = 1


class DNSCache(object):
    def __init__(self, loop, default_ttl=60):
        self._loop = loop or instance()
        self.default_ttl = default_ttl
        self._cache = defaultdict(list)
        self._last_resolve_time = time.time()

    def append(self, hostname, rrs):
        rrcs = []
        now = time.time()
        for rrc in self._cache[hostname]:
            if rrc.ttl_expried_time <= now:
                continue
            rrcs.append(rrc)
        self._cache[hostname] = rrcs

        for rr in rrs:
            has_cache = False
            for rrc in self._cache[hostname]:
                if rr == rrc:
                    has_cache = True
                    break
            if has_cache:
                continue
            setattr(rr, "ttl_expried_time", time.time() + (rr.ttl or self.default_ttl))
            self._cache[hostname].append(rr)

        if self._last_resolve_time - now >= 120:
            self._loop.add_async(self.resolve)
            self._last_resolve_time = now

    def get(self, hostname):
        if hostname not in self._cache or not self._cache[hostname]:
            return None, hostname

        now = time.time()
        if self._last_resolve_time - now >= 120:
            self._loop.add_async(self.resolve)
            self._last_resolve_time = now

        for rrc in self._cache[hostname]:
            if rrc.ttl_expried_time <= now:
                continue
            return str(rrc.rdata), hostname
        return None, hostname

    def remove(self, hostname):
        if hostname in self._cache:
            self._cache.pop(hostname)

        now = time.time()
        if self._last_resolve_time - now >= 120:
            self._loop.add_async(self.resolve)
            self._last_resolve_time = now

    def resolve(self):
        now = time.time()
        epried_hostnames = []
        for hostname, rrs in self._cache.items():
            rrcs = []
            for rr in rrs:
                if rr.ttl_expried_time <= now:
                    continue
                rrcs.append(rr)
            if rrcs:
                self._cache[hostname] = rrcs
            else:
                epried_hostnames.append(hostname)

        for hostname in epried_hostnames:
            self.remove(hostname)

    def clear(self):
        self._cache = defaultdict(list)

    def __getitem__(self, hostname):
        return self.get(hostname)[0]

    def __delitem__(self, hostname):
        return self.remove(hostname)

    def __contains__(self, hostname):
        return bool(self.get(hostname)[0])


class DnsQueryState(object):
    def __init__(self, hostname, v4server_count, v6server_count):
        self.hostname = hostname
        self.v4bv4_count = v4server_count
        self.v6bv6_count = v6server_count
        self.v6bv4_count = v4server_count
        self.v4bv6_count = v6server_count
        self.v4bv4_loading_count = 0
        self.v6bv6_loading_count = 0
        self.v6bv4_loading_count = 0
        self.v4bv6_loading_count = 0
        self.v4bv4_done_count = 0
        self.v6bv6_done_count = 0
        self.v6bv4_done_count = 0
        self.v4bv6_done_count = 0
        self.callbacks = []

    def v4bv4_done(self):
        return self.v4bv4_done_count >= self.v4bv4_count

    def v6bv6_done(self):
        return self.v6bv6_done_count >= self.v6bv6_count

    def v6bv4_done(self):
        return self.v6bv4_done_count >= self.v6bv4_count

    def v4bv6_done(self):
        return self.v4bv6_done_count >= self.v4bv6_count

    def done(self):
        return self.v4bv4_done_count >= self.v4bv4_count and self.v6bv6_done_count >= self.v6bv6_count \
               and self.v6bv4_done_count >= self.v6bv4_count and self.v4bv6_done_count >= self.v4bv6_count

    def append(self, callback):
        self.callbacks.append(callback)


class DNSResolver(EventEmitter):
    _instance = None

    @classmethod
    def default(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def __init__(self, loop=None, servers=None, hosts=None, resolve_timeout=None, resend_timeout=0.5):
        EventEmitter.__init__(self)

        self._loop = loop or instance()
        self._servers = []
        self._server6s = []
        self._hosts = hosts or {}

        self._cache = DNSCache(self._loop)
        self._queue = {}
        self._socket = None
        self._socket6 = None
        self._status = STATUS_OPENED

        if not servers:
            servers = self.parse_resolv()
        for server in servers:
            inet_type = self.is_ip(server)
            if inet_type == socket.AF_INET:
                self._servers.append(server)
            elif inet_type == socket.AF_INET6:
                self._server6s.append(server)

        if not hosts:
            self.parse_hosts()

        self._resolve_timeout = resolve_timeout if resolve_timeout else ((len(self._servers) + len(self._server6s))
                                                                         * resend_timeout + 4)
        self._resend_timeout = resend_timeout

    def on_resolve(self, callback):
        self.on("resolve", callback)

    def off_resolve(self, callback):
        self.on("resolve", callback)

    def once_resolve(self, callback):
        self.once("resolve", callback)

    def noce_resolve(self, callback):
        self.once("resolve", callback)

    def create_socket(self):
        from .udp import Socket
        self._socket = Socket(self._loop)
        self._socket.on_data(self.on_data)
        self._socket.on_close(self.on_close)
        self._socket.on_error(lambda s, e: None)

    def create_socket6(self):
        from .udp import Socket
        self._socket6 = Socket(self._loop)
        self._socket6.on_data(self.on_data6)
        self._socket6.on_close(self.on_close)
        self._socket6.on_error(lambda s, e: None)

    def parse_resolv(self):
        try:
            servers = [server for server in str(os.environ.get("SEVENT_NAMESERVER", '')).split(",")
                       if server and self.is_ip(server)]
            if servers:
                return servers
        except:
            pass

        servers = []
        etc_resolv_filename = os.environ.get("SEVENT_RESOLV_FILENAME")
        if etc_resolv_filename is None:
            etc_resolv_filename = '/etc/resolv.conf'
            if 'WINDIR' in os.environ:
                etc_resolv_filename = os.environ['WINDIR'] + '/system32/drivers/etc/resolv.conf'
        elif not etc_resolv_filename:
            return servers or ['8.8.4.4', '8.8.8.8', '114.114.114.114']
        try:
            with open(etc_resolv_filename, 'rb') as f:
                content = f.readlines()
                for line in content:
                    line = line.strip()
                    if not line or line[:1] == b'#' or not line.startswith(b'nameserver'):
                        continue

                    if is_py3 and type(line) != str:
                        parts = line.decode("utf-8").split()
                    else:
                        parts = line.split()
                    if len(parts) < 2:
                        continue
                    server = parts[1].strip()
                    if not self.is_ip(server):
                        continue
                    servers.append(server)
        except IOError:
            pass
        if not servers:
            if sys.platform == "win32":
                try:
                    from .win32util import get_dns_info
                    servers = get_dns_info().nameservers
                    if servers:
                        return servers
                except:
                    pass
        return servers or ['8.8.4.4', '8.8.8.8', '114.114.114.114']

    def parse_hosts(self):
        etc_hosts_filename = os.environ.get("SEVENT_HOSTS_FILENAME")
        if etc_hosts_filename is None:
            etc_hosts_filename = '/etc/hosts'
            if 'WINDIR' in os.environ:
                etc_hosts_filename = os.environ['WINDIR'] + '/system32/drivers/etc/hosts'
        elif not etc_hosts_filename or etc_hosts_filename == "-":
            self._hosts[b'localhost'] = '127.0.0.1'
            return
        try:
            with open(etc_hosts_filename, 'rb') as f:
                for line in f.readlines():
                    line = line.strip()
                    if not line or line[:1] == b'#':
                        continue

                    if is_py3 and type(line) != str:
                        parts = line.decode("utf-8").split()
                    else:
                        parts = line.split()
                    if len(parts) < 2:
                        continue
                    ip = parts[0].strip()
                    if not self.is_ip(ip):
                        continue

                    for i in range(1, len(parts)):
                        hostname = parts[i].strip()
                        if not hostname:
                            continue
                        self._hosts[ensure_bytes(hostname)] = ip
        except IOError:
            self._hosts[b'localhost'] = '127.0.0.1'

    def call_callback(self, hostname, ip):
        if hostname not in self._queue:
            return
        query_state = self._queue.pop(hostname)
        hostname = hostname.decode("utf-8") if is_py3 and type(hostname) != str else hostname
        for callback in query_state.callbacks:
            self._loop.add_async(callback, hostname, ip)
        self._loop.add_async(self.emit_resolve, self, hostname, ip)

    def on_data(self, socket, buffer):
        while buffer:
            data, address = buffer.next()
            try:
                answer = dnslib.DNSRecord.parse(data)
                hostname = b".".join(answer.q.qname.label)
                if hostname not in self._queue:
                    continue
                rrs = [rr for rr in answer.rr if rr.rtype == answer.q.qtype]
                query_state = self._queue[hostname]
                if answer.q.qtype == 28:
                    query_state.v6bv4_loading_count -= 1
                    query_state.v6bv4_done_count += 1
                else:
                    query_state.v4bv4_loading_count -= 1
                    query_state.v4bv4_done_count += 1

                if rrs:
                    self._cache.append(hostname, rrs)
                    self.call_callback(hostname, str(rrs[0].rdata))
                elif query_state.done():
                    self.call_callback(hostname, None)
                elif query_state.v4bv4_done() and query_state.v6bv4_loading_count <= 0 \
                        and query_state.v6bv4_done_count <= 0:
                    self.send_req(hostname, query_state, 0, True)
            except Exception:
                pass

    def on_data6(self, socket, buffer):
        while buffer:
            data, address = buffer.next()
            try:
                answer = dnslib.DNSRecord.parse(data)
                hostname = b".".join(answer.q.qname.label)
                if hostname not in self._queue:
                    continue
                rrs = [rr for rr in answer.rr if rr.rtype == answer.q.qtype]
                query_state = self._queue[hostname]
                if answer.q.qtype == 1:
                    query_state.v4bv6_loading_count -= 1
                    query_state.v4bv6_done_count += 1
                else:
                    query_state.v6bv6_loading_count -= 1
                    query_state.v6bv6_done_count += 1

                if rrs:
                    self._cache.append(hostname, rrs)
                    self.call_callback(hostname, str(rrs[0].rdata))
                elif query_state.done():
                    self.call_callback(hostname, None)
                elif query_state.v6bv6_done() and query_state.v4bv6_loading_count <= 0 \
                        and query_state.v4bv6_done_count <= 0:
                    self.send_req6(hostname, query_state, 0, True)
            except Exception:
                pass

    def send_req(self, hostname, query_state, server_index=0, is_query_v6=False):
        if not self._servers:
            return
        servers = self._servers

        question = dnslib.DNSRecord.question(hostname, 'AAAA' if is_query_v6 else 'A')
        if self._socket is None:
            self.create_socket()
        seq_data = bytes(question.pack())
        for _ in range(3):
            if server_index >= len(servers):
                break
            if is_query_v6:
                query_state.v6bv4_loading_count += 1
            else:
                query_state.v4bv4_loading_count += 1
            self._socket.write((seq_data, (servers[server_index], 53)))
            server_index += 1

        def on_timeout():
            if server_index >= len(servers):
                if hostname in self._queue and query_state.v6bv4_loading_count <= 0 \
                        and query_state.v6bv4_done_count <= 0:
                    self.send_req(hostname, query_state, 0, True)
                return
            if hostname in self._queue:
                self.send_req(hostname, query_state, server_index, is_query_v6)
        if server_index >= len(servers) and is_query_v6:
            return
        self._loop.add_timeout(self._resend_timeout, on_timeout)

    def send_req6(self, hostname, query_state, server_index=0, is_query_v4=False):
        if not self._server6s:
            return
        servers = self._server6s

        question = dnslib.DNSRecord.question(hostname, 'A' if is_query_v4 else 'AAAA')
        if self._socket6 is None:
            self.create_socket6()
        seq_data = bytes(question.pack())
        for _ in range(3):
            if server_index >= len(servers):
                break
            if is_query_v4:
                query_state.v4bv6_loading_count += 1
            else:
                query_state.v6bv6_loading_count += 1
            self._socket6.write((seq_data, (servers[server_index], 53)))
            server_index += 1

        def on_timeout():
            if server_index >= len(servers):
                if hostname in self._queue and query_state.v4bv6_loading_count <= 0 \
                        and query_state.v4bv6_done_count <= 0:
                    self.send_req6(hostname, query_state, 0, True)
                return
            if hostname in self._queue:
                self.send_req6(hostname, query_state, server_index, is_query_v4)
        if server_index >= len(servers) and is_query_v4:
            return
        self._loop.add_timeout(self._resend_timeout, on_timeout)

    def resolve(self, hostname, callback, timeout=None):
        if self._status == STATUS_CLOSED:
            return callback(hostname.decode("utf-8") if is_py3 and type(hostname) != str else hostname, None)

        hostname = ensure_bytes(hostname)
        if not hostname:
            return callback(hostname.decode("utf-8") if is_py3 and type(hostname) != str else hostname, None)
        elif self.is_ip(hostname):
            return callback(hostname.decode("utf-8") if is_py3 and type(hostname) != str else hostname, hostname)
        elif hostname in self._hosts:
            return callback(hostname.decode("utf-8") if is_py3 and type(hostname) != str else hostname, self._hosts[hostname])
        elif hostname in self._cache:
            return callback(hostname.decode("utf-8") if is_py3 and type(hostname) != str else hostname, self._cache[hostname])
        else:
            try:
                if hostname not in self._queue:
                    self._queue[hostname] = query_state = DnsQueryState(hostname, len(self._servers), len(self._server6s))
                    query_state.append(callback)
                    self.send_req(hostname, query_state)
                    self.send_req6(hostname, query_state)

                    if hostname in self._queue:
                        def on_timeout():
                            if hostname in self._queue:
                                self.call_callback(hostname, None)
                        self._loop.add_timeout(timeout or self._resolve_timeout, on_timeout)
                else:
                    self._queue[hostname].append(callback)
            except Exception:
                self.call_callback(hostname, None)
        return False

    def flush(self):
        self._cache.clear()

    def on_close(self, socket):
        if self._status == STATUS_CLOSED:
            return

        if self._socket == socket:
            self._socket = None
        else:
            self._socket6 = None

    def close(self):
        if self._status == STATUS_CLOSED:
            return

        self._status = STATUS_CLOSED
        if self._socket:
            self._socket.close()
        if self._socket6:
            self._socket6.close()

        for hostname, query_state in self._queue.items():
            for callback in query_state.callbacks:
                self._loop.add_async(callback, hostname, None)
        self._queue.clear()

    def is_ip(self, address):
        if is_py3 and type(address) != str:
            address = address.decode('utf8')

        try:
            socket.inet_pton(socket.AF_INET, address)
            return socket.AF_INET
        except (TypeError, ValueError, OSError, IOError):
            try:
                socket.inet_pton(socket.AF_INET6, address)
                return socket.AF_INET6
            except (TypeError, ValueError, OSError, IOError):
                return False


if is_py3:
    from .coroutines.dns import warp_coroutine
    DNSResolver = warp_coroutine(DNSResolver)


# -*- coding: utf-8 -*-
# 15/1/27
# create by: snower

import os
import time
from collections import deque
from .event import EventEmitter
from .loop import current
from .utils import get_logger

try:
    RECV_BUFFER_SIZE = int(os.environ.get("SEVENT_RECV_BUFFER_SIZE", 0))
except:
    RECV_BUFFER_SIZE = 0

try:
    MAX_BUFFER_SIZE = int(os.environ.get("SEVENT_MAX_BUFFER_SIZE", 4 * 1024 * 1024))
except:
    MAX_BUFFER_SIZE = 4 * 1024 * 1024

try:
    BUFFER_DRAIN_RATE = min(max(float(os.environ.get("SEVENT_BUFFER_DRAIN_RATE", 0.5)), 0.1), 0.9)
except:
    BUFFER_DRAIN_RATE = 0.5

try:
    RECV_COUNT = int(os.environ.get("SEVENT_RECV_COUNT", 0))
except:
    RECV_COUNT = 0
    
try:
    SEND_COUNT = int(os.environ.get("SEVENT_SEND_COUNT", 0))
except:
    SEND_COUNT = 0
    
try:
    if not os.environ.get("SEVENT_NOUSE_CBUFFER", False):
        from . import cbuffer
        BaseBuffer = cbuffer.Buffer
        if RECV_BUFFER_SIZE:
            try:
                cbuffer.socket_set_recv_size(RECV_BUFFER_SIZE)
            except: pass
        if RECV_COUNT:
            try:
                cbuffer.socket_set_recv_count(RECV_COUNT)
            except: pass
        if SEND_COUNT:
            try:
                cbuffer.socket_set_send_count(SEND_COUNT)
            except: pass
    else:
        cbuffer = None
except ImportError:
    get_logger().warning("cbuffer is not supported")
    cbuffer = None

RECV_BUFFER_SIZE = RECV_BUFFER_SIZE or 8 * 1024 - 64

if cbuffer is None:
    class BaseBuffer(object):
        def __init__(self):
            self._buffer = b''
            self._buffer_odata = None
            self._buffer_len = 0
            self._buffers = deque()
            self._buffers_odata = deque()
            self._len = 0
            self._buffer_index = 0

        def join(self):
            if self._buffer_index > 0 or self._buffer_len - self._buffer_index < self._len:
                if self._buffer_index < self._buffer_len:
                    self._buffers.appendleft(self._buffer[self._buffer_index:])
                if len(self._buffers) > 1:
                    data = b"".join(self._buffers)
                    self._buffers.clear()
                else:
                    data = self._buffers.popleft()
                if self._buffers_odata:
                    self._buffer_odata = self._buffers_odata[-1]
                    self._buffers_odata.clear()
                self._buffer = data
                self._buffer_len = len(data)
                self._buffer_index = 0
            return self._buffer

        def write(self, data, odata = None):
            if self._buffer_len > 0:
                self._buffers.append(data)
                self._buffers_odata.append(odata)
                self._len += len(data)
            else:
                self._buffer = data
                self._buffer_odata = odata
                self._buffer_len = len(data)
                self._len += self._buffer_len
            return self

        def read(self, size=-1):
            if size < 0:
                if self._buffer_len - self._buffer_index < self._len:
                    self.join()
                    buffer = self._buffer
                elif self._buffer_index > 0:
                    buffer = self._buffer[self._buffer_index:]
                else:
                    buffer = self._buffer
                buffer_odata = self._buffer_odata
                self._buffer_index, self._buffer_len, self._buffer, self._buffer_odata, self._len = 0, 0, b'', None, 0
                return (buffer, buffer_odata) if buffer_odata else buffer

            if self._len < size:
                return b""

            if self._buffer_len - self._buffer_index < size:
                self.join()
                data = self._buffer[:size]
                self._buffer_index = size
            else:
                data = self._buffer[self._buffer_index: self._buffer_index + size]
                self._buffer_index += size
            buffer_odata = self._buffer_odata
            self._len -= size

            if self._buffer_index >= self._buffer_len:
                if self._len > 0:
                    self._buffer = self._buffers.popleft()
                    self._buffer_odata = self._buffers_odata.popleft()
                    self._buffer_index, self._buffer_len = 0, len(self._buffer)
                else:
                    self._buffer_index, self._buffer_len, self._buffer, self._buffer_odata = 0, 0, b'', None
            return (data, buffer_odata) if buffer_odata else data

        def next(self):
            if self._buffer_index > 0:
                data = self._buffer[self._buffer_index:]
                self._len -= self._buffer_len - self._buffer_index
            else:
                data = self._buffer
                self._len -= self._buffer_len
            buffer_odata = self._buffer_odata

            if self._len > 0:
                self._buffer = self._buffers.popleft()
                self._buffer_odata = self._buffers_odata.popleft()
                self._buffer_index, self._buffer_len = 0, len(self._buffer)
            else:
                self._buffer_index, self._buffer_len, self._buffer, self._buffer_odata = 0, 0, b'', None
            return (data, buffer_odata) if buffer_odata else data

        def extend(self, o):
            if not isinstance(o, BaseBuffer):
                raise TypeError('not Buffer')

            while o:
                data = o.next()
                if isinstance(data, tuple):
                    self.write(*data)
                else:
                    self.write(data)

        def fetch(self, o, size=-1):
            if not isinstance(o, BaseBuffer):
                raise TypeError('not Buffer')

            data = o.read(size)
            self.write(data)
            return len(data)

        def copyfrom(self, o, size=-1):
            if not isinstance(o, BaseBuffer):
                raise TypeError('not Buffer')

            if size < 0:
                data = o.join()
            else:
                data = o.join()[:size]
            self.write(data)
            return len(data)

        def clear(self):
            self._buffer = b''
            self._buffer_odata = None
            self._buffer_len = 0
            self._buffers = deque()
            self._buffers_odata = deque()
            self._len = 0
            self._buffer_index = 0

        def head(self):
            if not self._buffer_odata:
                return self._buffer
            return (self._buffer, self._buffer_odata)

        def head_data(self):
            return self._buffer_odata

        def last(self):
            if not self._buffers:
                return self.head()

            if not self._buffers_odata[-1]:
                return self._buffers[-1]
            return (self._buffers[-1], self._buffers_odata[-1])

        def last_data(self):
            if not self._buffers_odata:
                return self.head_data()
            return self._buffers_odata[-1]

        def __len__(self):
            return self._len

        def __str__(self):
            buffer = self.join()
            return buffer.__str__()

        def __nonzero__(self):
            return self._len > 0

        def __getitem__(self, index):
            buffer = self.join()
            return (buffer.__getitem__(index), self._buffer_odata) if self._buffer_odata else buffer.__getitem__(index)

        def __hash__(self):
            buffer = self.join()
            return buffer.__hash__()


class Buffer(EventEmitter, BaseBuffer):
    def __init__(self, max_buffer_size=None):
        EventEmitter.__init__(self)
        BaseBuffer.__init__(self)

        self._loop = current()
        self._full = False
        self._drain_size = int(max_buffer_size or MAX_BUFFER_SIZE)
        self._regain_size = int(self._drain_size * BUFFER_DRAIN_RATE)
        self._drain_time = time.time()
        self._regain_time = time.time()

    @property
    def full(self):
        return self._full

    def on_drain(self, callback):
        self.on("drain", callback)

    def on_regain(self, callback):
        self.on("regain", callback)

    def once_drain(self, callback):
        self.once("drain", callback)

    def once_regain(self, callback):
        self.once("regain", callback)

    def _do_drain(self):
        self._full = True
        self._drain_time = time.time()
        try:
            self.emit_drain(self)
        except Exception as e:
            get_logger().exception("buffer emit drain error:%s", e)
    do_drain = _do_drain

    def _do_regain(self):
        self._full = False
        self._regain_time = time.time()
        try:
            self.emit_regain(self)
        except Exception as e:
            get_logger().exception("buffer emit regain error:%s", e)
    do_regain = _do_regain

    def write(self, data, odata=None):
        if odata is None:
            BaseBuffer.write(self, data)
        else:
            BaseBuffer.write(self, data, odata)

        if self._len > self._drain_size and not self._full:
            self.do_drain()
        return self

    def extend(self, o):
        BaseBuffer.extend(self, o)

        if o._full and o._len < o._regain_size:
            o.do_regain()

        if self._len > self._drain_size and not self._full:
            self.do_drain()
        return o

    def fetch(self, o, size=-1):
        r = BaseBuffer.fetch(self, o, size)

        if o._full and o._len < o._regain_size:
            o.do_regain()

        if self._len > self._drain_size and not self._full:
            self.do_drain()
        return r

    def copyfrom(self, o, size=-1):
        r = BaseBuffer.copyfrom(self, o, size)

        if self._len > self._drain_size and not self._full:
            self.do_drain()
        return r

    def read(self, size=-1):
        data = BaseBuffer.read(self, size)

        if self._full and self._len < self._regain_size:
            self.do_regain()
        return data

    def next(self):
        data = BaseBuffer.next(self)

        if self._full and self._len < self._regain_size:
            self.do_regain()
        return data

    def clear(self):
        BaseBuffer.clear(self)

        if self._full and self._len < self._regain_size:
            self.do_regain()

    def link(self, o):
        self_do_drain, self_do_regain = self.do_drain, self.do_regain
        o_do_drain, o_do_regain = o.do_drain, o.do_regain

        def do_drain():
            self_do_drain()
            o_do_drain()

        def do_regain():
            if self._len >= self._regain_size or o._len >= o._regain_size:
                return
            self_do_regain()
            o_do_regain()

        self.do_drain, self.do_regain = do_drain, do_regain
        o.do_drain, o.do_regain = do_drain, do_regain
        if (self._full and not o._full) or (not self._full and o._full):
            do_drain()
        return self

    def close(self):
        self.do_drain = self._do_drain
        self.do_regain = self._do_regain
        self.remove_all_listeners()

    def decode(self, *args, **kwargs):
        data = self.join()
        return data.decode(*args, **kwargs)

    def items(self):
        return self._buffers

    def __getitem__(self, item):
        data = self.join()
        return data.__getitem__(item)

    def __iter__(self):
        data = self.join()
        return iter(data)

    def __contains__(self, item):
        data = self.join()

        return data.__contains__(item)

    def __add__(self, other):
        data = self.join()
        return data.__add__(other)

    def __bytes__(self):
        return self.join()

    def __cmp__(self, other):
        data = self.join()
        return data.__cmp__(other)

    def __eq__(self, other):
        data = self.join()
        return data == other

    def __gt__(self, other):
        data = self.join()
        return data > other

    def __lt__(self, other):
        data = self.join()
        return data < other

    def __ge__(self, other):
        data = self.join()
        return data >= other

    def __le__(self, other):
        data = self.join()
        return data <= other

    def __ne__(self, other):
        data = self.join()
        return data != other
